import operator
import warnings

from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import cached_property
from typing import Literal

import arviz as az
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import numpyro

from jax import random
from jax.random import PRNGKey
from numpyro.contrib.nested_sampling import NestedSampler
from numpyro.distributions import Poisson, TransformedDistribution
from numpyro.infer import AIES, ESS, MCMC, NUTS, Predictive
from numpyro.infer.inspect import get_model_relations
from numpyro.infer.reparam import TransformReparam
from numpyro.infer.util import log_density

from ._fit._build_model import build_prior, forward_model
from .analysis._plot import _plot_poisson_data_with_error
from .analysis.results import FitResult
from .data import ObsConfiguration
from .model.abc import SpectralModel
from .model.background import BackgroundModel
from .util.typing import PriorDictModel, PriorDictType


class BayesianModel:
    """
    Base class for a Bayesian model. This class contains the necessary methods to build a model, sample from the prior
    and compute the log-likelihood and posterior probability.
    """

    def __init__(
        self,
        model: SpectralModel,
        prior_distributions: PriorDictType | Callable,
        observations: ObsConfiguration | list[ObsConfiguration] | dict[str, ObsConfiguration],
        background_model: BackgroundModel = None,
        sparsify_matrix: bool = False,
    ):
        """
        Build a Bayesian model for a given spectral model and observations.

        Parameters:
            model: the spectral model to fit.
            prior_distributions: a nested dictionary containing the prior distributions for the model parameters, or a
                callable function that returns parameter samples.
            observations: the observations to fit the model to.
            background_model: the background model to fit.
            sparsify_matrix: whether to sparsify the transfer matrix.
        """
        self.model = model
        self._observations = observations
        self.background_model = background_model
        self.sparse = sparsify_matrix

        if not callable(prior_distributions):
            # Validate the entry with pydantic
            prior = PriorDictModel.from_dict(prior_distributions).nested_dict

            def prior_distributions_func():
                return build_prior(prior, expand_shape=(len(self.observation_container),))

        else:
            prior_distributions_func = prior_distributions

        self.prior_distributions_func = prior_distributions_func
        self.init_params = self.prior_samples()

    @cached_property
    def observation_container(self) -> dict[str, ObsConfiguration]:
        """
        The observations used in the fit as a dictionary of observations.
        """

        if isinstance(self._observations, dict):
            return self._observations

        elif isinstance(self._observations, list):
            return {f"data_{i}": obs for i, obs in enumerate(self._observations)}

        elif isinstance(self._observations, ObsConfiguration):
            return {"data": self._observations}

        else:
            raise ValueError(f"Invalid type for observations : {type(self._observations)}")

    @cached_property
    def numpyro_model(self) -> Callable:
        """
        Build the numpyro model using the observed data, the prior distributions and the spectral model.
        """

        def numpyro_model(observed=True):
            # Instantiate and register the parameters of the spectral model and the background
            prior_params = self.prior_distributions_func()

            # Iterate over all the observations in our container and build a single numpyro model for each observation
            for i, (name, observation) in enumerate(self.observation_container.items()):
                # Check that we can indeed fit a background
                if (getattr(observation, "folded_background", None) is not None) and (
                    self.background_model is not None
                ):
                    # This call should register the parameter and observation of our background model
                    bkg_countrate = self.background_model.numpyro_model(
                        observation, name=name, observed=observed
                    )

                elif (getattr(observation, "folded_background", None) is None) and (
                    self.background_model is not None
                ):
                    raise ValueError(
                        "Trying to fit a background model but no background is linked to this observation"
                    )

                else:
                    bkg_countrate = 0.0

                # We expect that prior_params contains an array of parameters for each observation
                # They can be identical or different for each observation
                params = jax.tree.map(lambda x: x[i], prior_params)

                # Forward model the observation and get the associated countrate
                obs_model = jax.jit(
                    lambda par: forward_model(self.model, par, observation, sparse=self.sparse)
                )
                obs_countrate = obs_model(params)

                # Register the observation as an observed site
                with numpyro.plate("obs_plate_" + name, len(observation.folded_counts)):
                    numpyro.sample(
                        "obs_" + name,
                        Poisson(obs_countrate + bkg_countrate / observation.folded_backratio.data),
                        obs=observation.folded_counts.data if observed else None,
                    )

        return numpyro_model

    @cached_property
    def transformed_numpyro_model(self) -> Callable:
        transform_dict = {}

        relations = get_model_relations(self.numpyro_model)
        distributions = {
            parameter: getattr(numpyro.distributions, value, None)
            for parameter, value in relations["sample_dist"].items()
        }

        for parameter, distribution in distributions.items():
            if isinstance(distribution, TransformedDistribution):
                transform_dict[parameter] = TransformReparam()

        return numpyro.handlers.reparam(self.numpyro_model, config=transform_dict)

    @cached_property
    def log_likelihood_per_obs(self) -> Callable:
        """
        Build the log likelihood function for each bins in each observation.
        """

        @jax.jit
        def log_likelihood_per_obs(constrained_params):
            log_likelihood = numpyro.infer.util.log_likelihood(
                model=self.numpyro_model, posterior_samples=constrained_params
            )

            return jax.tree.map(lambda x: jnp.where(jnp.isnan(x), -jnp.inf, x), log_likelihood)

        return log_likelihood_per_obs

    @cached_property
    def log_likelihood(self) -> Callable:
        """
        Build the total log likelihood function. Takes a dictionary of parameters where the keys are the parameter names
        that can be fetched with the [`parameter_names`][jaxspec.fit.BayesianModel.parameter_names].
        """

        @jax.jit
        def log_likelihood(constrained_params):
            log_likelihood = self.log_likelihood_per_obs(constrained_params)

            return jax.tree.reduce(operator.add, jax.tree.map(jnp.sum, log_likelihood))

        return log_likelihood

    @cached_property
    def log_posterior_prob(self) -> Callable:
        """
        Build the posterior probability. Takes a dictionary of parameters where the keys are the parameter names
        that can be fetched with the [`parameter_names`][jaxspec.fit.BayesianModel.parameter_names].
        """

        # This is required as numpyro.infer.util.log_densities does not check parameter validity by itself
        numpyro.enable_validation()

        @jax.jit
        def log_posterior_prob(constrained_params):
            log_posterior_prob, _ = log_density(
                self.numpyro_model, (), dict(observed=True), constrained_params
            )
            return jnp.where(jnp.isnan(log_posterior_prob), -jnp.inf, log_posterior_prob)

        return log_posterior_prob

    @cached_property
    def parameter_names(self) -> list[str]:
        """
        A list of parameter names for the model.
        """
        relations = get_model_relations(self.numpyro_model)
        all_sites = relations["sample_sample"].keys()
        observed_sites = relations["observed"]
        return [site for site in all_sites if site not in observed_sites]

    @cached_property
    def observation_names(self) -> list[str]:
        """
        List of the observations.
        """
        relations = get_model_relations(self.numpyro_model)
        all_sites = relations["sample_sample"].keys()
        observed_sites = relations["observed"]
        return [site for site in all_sites if site in observed_sites]

    def array_to_dict(self, theta):
        """
        Convert an array of parameters to a dictionary of parameters.
        """
        input_params = {}

        for index, key in enumerate(self.parameter_names):
            input_params[key] = theta[index]

        return input_params

    def dict_to_array(self, dict_of_params):
        """
        Convert a dictionary of parameters to an array of parameters.
        """

        theta = jnp.zeros(len(self.parameter_names))

        for index, key in enumerate(self.parameter_names):
            theta = theta.at[index].set(dict_of_params[key])

        return theta

    def prior_samples(self, key: PRNGKey = PRNGKey(0), num_samples: int = 100):
        """
        Get initial parameters for the model by sampling from the prior distribution

        Parameters:
            key: the random key used to initialize the sampler.
            num_samples: the number of samples to draw from the prior.
        """

        @jax.jit
        def prior_sample(key):
            return Predictive(
                self.numpyro_model, return_sites=self.parameter_names, num_samples=num_samples
            )(key, observed=False)

        return prior_sample(key)

    def mock_observations(self, parameters, key: PRNGKey = PRNGKey(0)):
        @jax.jit
        def fakeit(key, parameters):
            return Predictive(
                self.numpyro_model,
                return_sites=self.observation_names,
                posterior_samples=parameters,
            )(key, observed=False)

        return fakeit(key, parameters)

    def prior_predictive_coverage(
        self,
        key: PRNGKey = PRNGKey(0),
        num_samples: int = 1000,
    ):
        """
        Check if the prior distribution include the observed data.
        """
        key_prior, key_posterior = jax.random.split(key, 2)
        prior_params = self.prior_samples(key=key_prior, num_samples=num_samples)
        posterior_observations = self.mock_observations(prior_params, key=key_posterior)

        for key, value in self.observation_container.items():
            fig, axs = plt.subplots(
                nrows=2, ncols=1, sharex=True, figsize=(5, 6), height_ratios=[3, 1]
            )

            _plot_poisson_data_with_error(
                axs[0],
                value.out_energies,
                value.folded_counts.values,
                percentiles=(16, 84),
            )

            for i, (envelop_percentiles, color, alpha) in enumerate(
                zip(
                    [(16, 86), (2.5, 97.5), (0.15, 99.85)],
                    ["#03045e", "#0077b6", "#00b4d8"],
                    [0.5, 0.4, 0.3],
                )
            ):
                lower, upper = np.percentile(
                    posterior_observations["obs_" + key], envelop_percentiles, axis=0
                )

                axs[0].stairs(
                    upper,
                    edges=[*list(value.out_energies[0]), value.out_energies[1][-1]],
                    baseline=lower,
                    alpha=alpha,
                    fill=True,
                    color=color,
                    label=rf"${1+i}\sigma$",
                )

            # rank = np.vstack((posterior_observations["obs_" + key], value.folded_counts.values)).argsort(axis=0)[-1] / (num_samples) * 100
            counts = posterior_observations["obs_" + key]
            observed = value.folded_counts.values

            num_samples = counts.shape[0]

            less_than_obs = (counts < observed).sum(axis=0)
            equal_to_obs = (counts == observed).sum(axis=0)

            rank = (less_than_obs + 0.5 * equal_to_obs) / num_samples * 100

            axs[1].stairs(rank, edges=[*list(value.out_energies[0]), value.out_energies[1][-1]])

            axs[1].plot(
                (value.out_energies.min(), value.out_energies.max()),
                (50, 50),
                color="black",
                linestyle="--",
            )

            axs[1].set_xlabel("Energy (keV)")
            axs[0].set_ylabel("Counts")
            axs[1].set_ylabel("Rank (%)")
            axs[1].set_ylim(0, 100)
            axs[0].set_xlim(value.out_energies.min(), value.out_energies.max())
            axs[0].loglog()
            axs[0].legend(loc="upper right")
            plt.suptitle(f"Prior Predictive coverage for {key}")
            plt.tight_layout()
            plt.show()


class BayesianModelFitter(BayesianModel, ABC):
    def build_inference_data(
        self,
        posterior_samples,
        num_chains: int = 1,
        num_predictive_samples: int = 1000,
        key: PRNGKey = PRNGKey(42),
        use_transformed_model: bool = False,
        filter_inference_data: bool = True,
    ) -> az.InferenceData:
        """
        Build an [InferenceData][arviz.InferenceData] object from posterior samples.

        Parameters:
            posterior_samples: the samples from the posterior distribution.
            num_chains: the number of chains used to sample the posterior.
            num_predictive_samples: the number of samples to draw from the prior.
            key: the random key used to initialize the sampler.
            use_transformed_model: whether to use the transformed model to build the InferenceData.
            filter_inference_data: whether to filter the InferenceData to keep only the relevant parameters.
        """

        numpyro_model = (
            self.transformed_numpyro_model if use_transformed_model else self.numpyro_model
        )

        keys = random.split(key, 3)

        posterior_predictive = Predictive(numpyro_model, posterior_samples)(keys[0], observed=False)

        prior = Predictive(numpyro_model, num_samples=num_predictive_samples * num_chains)(
            keys[1], observed=False
        )

        log_likelihood = numpyro.infer.log_likelihood(numpyro_model, posterior_samples)

        seeded_model = numpyro.handlers.substitute(
            numpyro.handlers.seed(numpyro_model, keys[3]),
            substitute_fn=numpyro.infer.init_to_sample,
        )

        observations = {
            name: site["value"]
            for name, site in numpyro.handlers.trace(seeded_model).get_trace().items()
            if site["type"] == "sample" and site["is_observed"]
        }

        def reshape_first_dimension(arr):
            new_dim = arr.shape[0] // num_chains
            new_shape = (num_chains, new_dim) + arr.shape[1:]
            reshaped_array = arr.reshape(new_shape)

            return reshaped_array

        posterior_samples = {
            key: reshape_first_dimension(value) for key, value in posterior_samples.items()
        }
        prior = {key: value[None, :] for key, value in prior.items()}
        posterior_predictive = {
            key: reshape_first_dimension(value) for key, value in posterior_predictive.items()
        }
        log_likelihood = {
            key: reshape_first_dimension(value) for key, value in log_likelihood.items()
        }

        inference_data = az.from_dict(
            posterior_samples,
            prior=prior,
            posterior_predictive=posterior_predictive,
            log_likelihood=log_likelihood,
            observed_data=observations,
        )

        return (
            self.filter_inference_data(inference_data) if filter_inference_data else inference_data
        )

    def filter_inference_data(
        self,
        inference_data: az.InferenceData,
    ) -> az.InferenceData:
        """
        Filter the inference data to keep only the relevant parameters for the observations.

        - Removes predictive parameters from deterministic random variables (e.g. kernel of background GP)
        - Removes parameters build from reparametrised variables (e.g. ending with `"_base"`)
        """

        predictive_parameters = []

        for key, value in self.observation_container.items():
            if self.background_model is not None:
                predictive_parameters.append(f"obs_{key}")
                predictive_parameters.append(f"bkg_{key}")
            else:
                predictive_parameters.append(f"obs_{key}")

        inference_data.posterior_predictive = inference_data.posterior_predictive[
            predictive_parameters
        ]

        parameters = [
            x
            for x in inference_data.posterior.keys()
            if not x.endswith("_base") or x.startswith("_")
        ]
        inference_data.posterior = inference_data.posterior[parameters]
        inference_data.prior = inference_data.prior[parameters]

        return inference_data

    @abstractmethod
    def fit(self, **kwargs) -> FitResult: ...


class MCMCFitter(BayesianModelFitter):
    """
    A class to fit a model to a given set of observation using a Bayesian approach. This class uses samplers
    from numpyro to perform the inference on the model parameters.
    """

    kernel_dict = {
        "nuts": NUTS,
        "aies": AIES,
        "ess": ESS,
    }

    def fit(
        self,
        rng_key: int = 0,
        num_chains: int = len(jax.devices()),
        num_warmup: int = 1000,
        num_samples: int = 1000,
        sampler: Literal["nuts", "aies", "ess"] = "nuts",
        use_transformed_model: bool = True,
        kernel_kwargs: dict = {},
        mcmc_kwargs: dict = {},
    ) -> FitResult:
        """
        Fit the model to the data using a MCMC sampler from numpyro.

        Parameters:
            rng_key: the random key used to initialize the sampler.
            num_chains: the number of chains to run.
            num_warmup: the number of warmup steps.
            num_samples: the number of samples to draw.
            sampler: the sampler to use. Can be one of "nuts", "aies" or "ess".
            use_transformed_model: whether to use the transformed model to build the InferenceData.
            kernel_kwargs: additional arguments to pass to the kernel. See [`NUTS`][numpyro.infer.mcmc.MCMCKernel] for more details.
            mcmc_kwargs: additional arguments to pass to the MCMC sampler. See [`MCMC`][numpyro.infer.mcmc.MCMC] for more details.

        Returns:
            A [`FitResult`][jaxspec.analysis.results.FitResult] instance containing the results of the fit.
        """

        bayesian_model = (
            self.transformed_numpyro_model if use_transformed_model else self.numpyro_model
        )

        chain_kwargs = {
            "num_warmup": num_warmup,
            "num_samples": num_samples,
            "num_chains": num_chains,
        }

        kernel = self.kernel_dict[sampler](bayesian_model, **kernel_kwargs)

        mcmc_kwargs = chain_kwargs | mcmc_kwargs

        if sampler in ["aies", "ess"] and mcmc_kwargs.get("chain_method", None) != "vectorized":
            mcmc_kwargs["chain_method"] = "vectorized"
            warnings.warn("The chain_method is set to 'vectorized' for AIES and ESS samplers")

        mcmc = MCMC(kernel, **mcmc_kwargs)
        keys = random.split(random.PRNGKey(rng_key), 3)

        mcmc.run(keys[0])

        posterior = mcmc.get_samples()

        inference_data = self.build_inference_data(
            posterior, num_chains=num_chains, use_transformed_model=True
        )

        return FitResult(
            self,
            inference_data,
            self.model.params,
            background_model=self.background_model,
        )


class NSFitter(BayesianModelFitter):
    r"""
    A class to fit a model to a given set of observation using the Nested Sampling algorithm. This class uses the
    [`DefaultNestedSampler`][jaxns.DefaultNestedSampler] from [`jaxns`](https://jaxns.readthedocs.io/en/latest/) which
    implements the [Phantom-Powered Nested Sampling](https://arxiv.org/abs/2312.11330) algorithm.

    !!! info
        Ensure large prior volume is covered by the prior distributions to ensure the algorithm yield proper results.

    """

    def fit(
        self,
        rng_key: int = 0,
        num_samples: int = 1000,
        num_live_points: int = 1000,
        plot_diagnostics=False,
        termination_kwargs: dict | None = None,
        verbose=True,
    ) -> FitResult:
        """
        Fit the model to the data using the Phantom-Powered nested sampling algorithm.

        Parameters:
            rng_key: the random key used to initialize the sampler.
            num_samples: the number of samples to draw.
            num_live_points: the number of live points to use at the start of the NS algorithm.
            plot_diagnostics: whether to plot the diagnostics of the NS algorithm.
            termination_kwargs: additional arguments to pass to the termination criterion of the NS algorithm.
            verbose: whether to print the progress of the NS algorithm.

        Returns:
            A [`FitResult`][jaxspec.analysis.results.FitResult] instance containing the results of the fit.
        """

        bayesian_model = self.transformed_numpyro_model
        keys = random.split(random.PRNGKey(rng_key), 4)

        ns = NestedSampler(
            bayesian_model,
            constructor_kwargs=dict(
                num_parallel_workers=1,
                verbose=verbose,
                difficult_model=True,
                max_samples=1e6,
                parameter_estimation=True,
                num_live_points=num_live_points,
            ),
            termination_kwargs=termination_kwargs if termination_kwargs else dict(),
        )

        ns.run(keys[0])

        if plot_diagnostics:
            ns.diagnostics()

        posterior = ns.get_samples(keys[1], num_samples=num_samples)
        inference_data = self.build_inference_data(
            posterior, num_chains=1, use_transformed_model=True
        )

        return FitResult(
            self,
            inference_data,
            self.model.params,
            background_model=self.background_model,
        )



import numpy as np
from .observation import Observation


def constant_rebin(obs: Observation, downfactor: int):
    """
    Create a grouping matrix for constant rebinning of the data.

    Parameters:
        obs: Observation object.
        downfactor: Downfactor for the rebinning.

    Returns:
        A grouping matrix.
    """
    n_channels = len(obs.pha.channel)
    n_bins = n_channels // downfactor + 1 * (n_channels % downfactor != 0)
    grouping = np.zeros((n_bins, n_channels), dtype=bool)

    for i in range(n_bins):
        grouping[i, i * downfactor : (i + 1) * downfactor] = True

    return grouping



from collections.abc import Mapping
from pathlib import Path
from typing import Literal, TypeVar

import haiku as hk
import jax
import numpy as np
import numpyro

from astropy.io import fits
from numpy.typing import ArrayLike
from numpyro import handlers

from .._fit._build_model import CountForwardModel
from ..model.abc import SpectralModel
from ..util.online_storage import table_manager
from . import Instrument, ObsConfiguration, Observation

K = TypeVar("K")
V = TypeVar("V")


def load_example_pha(
    source: Literal["NGC7793_ULX4_PN", "NGC7793_ULX4_ALL"],
) -> (Observation, list[Observation] | dict[str, Observation]):
    """
    Load some example observations from the package data.

    Parameters:
        source: The source to be loaded. Can be either "NGC7793_ULX4_PN" or "NGC7793_ULX4_ALL".
    """

    if source == "NGC7793_ULX4_PN":
        return Observation.from_pha_file(
            table_manager.fetch("example_data/NGC7793_ULX4/PN_spectrum_grp20.fits"),
            bkg_path=table_manager.fetch("example_data/NGC7793_ULX4/PNbackground_spectrum.fits"),
        )

    elif source == "NGC7793_ULX4_ALL":
        return {
            "PN": Observation.from_pha_file(
                table_manager.fetch("example_data/NGC7793_ULX4/PN_spectrum_grp20.fits"),
                bkg_path=table_manager.fetch(
                    "example_data/NGC7793_ULX4/PNbackground_spectrum.fits"
                ),
            ),
            "MOS1": Observation.from_pha_file(
                table_manager.fetch("example_data/NGC7793_ULX4/MOS1_spectrum_grp.fits"),
                bkg_path=table_manager.fetch(
                    "example_data/NGC7793_ULX4/MOS1background_spectrum.fits"
                ),
            ),
            "MOS2": Observation.from_pha_file(
                table_manager.fetch("example_data/NGC7793_ULX4/MOS2_spectrum_grp.fits"),
                bkg_path=table_manager.fetch(
                    "example_data/NGC7793_ULX4/MOS2background_spectrum.fits"
                ),
            ),
        }

    else:
        raise ValueError(f"{source} not recognized.")


def load_example_instruments(source: Literal["NGC7793_ULX4_PN", "NGC7793_ULX4_ALL"]):
    """
    Load some example instruments from the package data.

    Parameters:
        source: The source to be loaded. Can be either "NGC7793_ULX4_PN" or "NGC7793_ULX4_ALL".

    """
    if source == "NGC7793_ULX4_PN":
        return Instrument.from_ogip_file(
            table_manager.fetch("example_data/NGC7793_ULX4/PN.rmf"),
            table_manager.fetch("example_data/NGC7793_ULX4/PN.arf"),
        )

    elif source == "NGC7793_ULX4_ALL":
        return {
            "PN": Instrument.from_ogip_file(
                table_manager.fetch("example_data/NGC7793_ULX4/PN.rmf"),
                table_manager.fetch("example_data/NGC7793_ULX4/PN.arf"),
            ),
            "MOS1": Instrument.from_ogip_file(
                table_manager.fetch("example_data/NGC7793_ULX4/MOS1.rmf"),
                table_manager.fetch("example_data/NGC7793_ULX4/MOS1.arf"),
            ),
            "MOS2": Instrument.from_ogip_file(
                table_manager.fetch("example_data/NGC7793_ULX4/MOS2.rmf"),
                table_manager.fetch("example_data/NGC7793_ULX4/MOS2.arf"),
            ),
        }

    else:
        raise ValueError(f"{source} not recognized.")


def load_example_obsconf(source: Literal["NGC7793_ULX4_PN", "NGC7793_ULX4_ALL"]):
    """
    Load some example ObsConfigurations.

    Parameters:
        source: The source to be loaded. Can be either "NGC7793_ULX4_PN" or "NGC7793_ULX4_ALL".
    """

    if source in "NGC7793_ULX4_PN":
        instrument = load_example_instruments(source)
        observation = load_example_pha(source)

        return ObsConfiguration.from_instrument(
            instrument, observation, low_energy=0.5, high_energy=8.0
        )

    elif source == "NGC7793_ULX4_ALL":
        instruments_dict = load_example_instruments(source)
        observations_dict = load_example_pha(source)

        return {
            key: ObsConfiguration.from_instrument(
                instruments_dict[key], observations_dict[key], low_energy=0.5, high_energy=8.0
            )
            for key in instruments_dict.keys()
        }

    else:
        raise ValueError(f"{source} not recognized.")


def fakeit(
    instrument: ObsConfiguration | list[ObsConfiguration],
    model: SpectralModel,
    parameters: Mapping[K, V],
    rng_key: int = 0,
    sparsify_matrix: bool = False,
) -> ArrayLike | list[ArrayLike]:
    """
    Convenience function to simulate a spectrum from a given model and a set of parameters.
    It requires an instrumental setup, and unlike in
    [XSPEC's fakeit](https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/node72.html), the error on the counts is given
    exclusively by Poisson statistics.

    Parameters:
        instrument: The instrumental setup.
        model: The model to use.
        parameters: The parameters of the model.
        rng_key: The random number generator seed.
        sparsify_matrix: Whether to sparsify the matrix or not.
    """

    instruments = [instrument] if isinstance(instrument, ObsConfiguration) else instrument
    fakeits = []

    for i, instrument in enumerate(instruments):
        transformed_model = hk.without_apply_rng(
            hk.transform(
                lambda par: CountForwardModel(model, instrument, sparse=sparsify_matrix)(par)
            )
        )

        def obs_model(p):
            return transformed_model.apply(None, p)

        with handlers.seed(rng_seed=rng_key):
            counts = numpyro.sample(
                f"likelihood_obs_{i}",
                numpyro.distributions.Poisson(obs_model(parameters)),
            )

        """
        pha = DataPHA(
            instrument.rmf.channel,
            np.array(counts, dtype=int)*u.ct,
            instrument.exposure,
            grouping=instrument.grouping)

        observation = Observation(
            pha=pha,
            arf=instrument.arf,
            rmf=instrument.rmf,
            low_energy=instrument.low_energy,
            high_energy=instrument.high_energy
        )
        """

        fakeits.append(np.array(counts, dtype=int))

    return fakeits[0] if len(fakeits) == 1 else fakeits


def fakeit_for_multiple_parameters(
    instrument: ObsConfiguration | list[ObsConfiguration],
    model: SpectralModel,
    parameters: Mapping[K, V],
    rng_key: int = 0,
    apply_stat: bool = True,
    sparsify_matrix: bool = False,
):
    """
    Convenience function to simulate multiple spectra from a given model and a set of parameters.

    TODO : avoid redundancy, better doc and type hints

    Parameters:
        instrument: The instrumental setup.
        model: The model to use.
        parameters: The parameters of the model.
        rng_key: The random number generator seed.
        apply_stat: Whether to apply Poisson statistic on the folded spectra or not.
        sparsify_matrix: Whether to sparsify the matrix or not.
    """

    instruments = [instrument] if isinstance(instrument, ObsConfiguration) else instrument
    fakeits = []

    for i, obs in enumerate(instruments):
        transformed_model = hk.without_apply_rng(
            hk.transform(lambda par: CountForwardModel(model, obs, sparse=sparsify_matrix)(par))
        )

        @jax.jit
        @jax.vmap
        def obs_model(p):
            return transformed_model.apply(None, p)

        if apply_stat:
            with handlers.seed(rng_seed=rng_key):
                spectrum = numpyro.sample(
                    f"likelihood_obs_{i}",
                    numpyro.distributions.Poisson(obs_model(parameters)),
                )

        else:
            spectrum = obs_model(parameters)

        fakeits.append(spectrum)

    return fakeits[0] if len(fakeits) == 1 else fakeits


def data_path_finder(pha_path: str) -> tuple[str | None, str | None, str | None]:
    """
    Function which tries its best to find the ARF, RMF and BKG files associated with a given PHA file.

    Parameters:
        pha_path: The PHA file path.

    Returns:
        arf_path: The ARF file path.
        rmf_path: The RMF file path.
        bkg_path: The BKG file path.
    """

    def find_path(file_name: str, directory: str) -> str | None:
        if file_name.lower() != "none" and file_name != "":
            return find_file_or_compressed_in_dir(file_name, directory)
        else:
            return None

    header = fits.getheader(pha_path, "SPECTRUM")
    directory = str(Path(pha_path).parent)

    arf_path = find_path(header.get("ANCRFILE", "none"), directory)
    rmf_path = find_path(header.get("RESPFILE", "none"), directory)
    bkg_path = find_path(header.get("BACKFILE", "none"), directory)

    return arf_path, rmf_path, bkg_path


def find_file_or_compressed_in_dir(path: str | Path, directory: str | Path) -> str:
    """
    Try to find a file or its .gz compressed version in a given directory and return
    the full path of the file.
    """
    path = Path(path) if isinstance(path, str) else path
    directory = Path(directory) if isinstance(directory, str) else directory

    if directory.joinpath(path).exists():
        return str(directory.joinpath(path))

    matching_files = list(directory.glob(str(path) + "*"))

    if matching_files:
        file = matching_files[0]
        if file.suffix == ".gz":
            return str(file)

    else:
        raise FileNotFoundError(f"Can't find {path}(.gz) in {directory}.")



from typing import Any

import numpyro.distributions as dist

from jax import numpy as jnp
from jax.typing import ArrayLike
from pydantic import BaseModel, field_validator

PriorDictType = dict[str, dict[str, dist.Distribution | ArrayLike]]


def is_flat_dict(input_data: dict[str, Any]) -> bool:
    """
    Check if the input data is a flat dictionary with string keys and non-dictionary values.
    """
    return all(isinstance(k, str) and not isinstance(v, dict) for k, v in input_data.items())


class PriorDictModel(BaseModel):
    """
    Pydantic model for a nested dictionary of NumPyro distributions or JAX arrays.
    The top level keys are strings, and the values are dictionaries with string keys and values that are either
    NumPyro distributions or JAX arrays (or convertible to JAX arrays).
    """

    nested_dict: PriorDictType

    class Config:  # noqa D106
        arbitrary_types_allowed = True

    @classmethod
    def from_dict(cls, input_prior: dict[str, Any]):
        if is_flat_dict(input_prior):
            nested_dict = {}

            for key, obj in input_prior.items():
                component, component_number, *parameter = key.split("_")

                sub_dict = nested_dict.get(f"{component}_{component_number}", {})
                sub_dict["_".join(parameter)] = obj

                nested_dict[f"{component}_{component_number}"] = sub_dict

            return cls(nested_dict=nested_dict)

        return cls(nested_dict=input_prior)

    @field_validator("nested_dict", mode="before")
    def check_and_cast_nested_dict(cls, value: dict[str, Any]):
        if not isinstance(value, dict):
            raise ValueError("The top level must be a dictionary")

        for key, inner_dict in value.items():
            if not isinstance(inner_dict, dict):
                raise ValueError(f'The value for key "{key}" must be a dictionary')

            for inner_key, obj in inner_dict.items():
                if not isinstance(obj, dist.Distribution):
                    try:
                        # Attempt to cast to JAX array
                        value[key][inner_key] = jnp.array(obj, dtype=float)

                    except Exception as e:
                        raise ValueError(
                            f'The value for key "{inner_key}" in {key} be a NumPyro '
                            f"distribution or castable to JAX array. Error: {e}"
                        )
        return value



import matplotlib.pyplot as plt
import numpy as np

from jax.typing import ArrayLike
from scipy.stats import nbinom


def _plot_poisson_data_with_error(
    ax: plt.Axes,
    x_bins: ArrayLike,
    y: ArrayLike,
    percentiles: tuple = (16, 84),
):
    """
    Plot Poisson data with error bars. We extrapolate the intrinsic error of the observation assuming a prior rate
    distributed according to a Gamma RV.
    """
    y_low = nbinom.ppf(percentiles[0] / 100, y, 0.5)
    y_high = nbinom.ppf(percentiles[1] / 100, y, 0.5)

    ax_to_plot = ax.errorbar(
        np.sqrt(x_bins[0] * x_bins[1]),
        y,
        xerr=np.abs(x_bins - np.sqrt(x_bins[0] * x_bins[1])),
        yerr=[
            y - y_low,
            y_high - y,
        ],
        color="black",
        linestyle="none",
        alpha=0.3,
        capsize=2,
    )

    return ax_to_plot



import os
import numpy as np
import xarray as xr
from matplotlib import colors
from .ogip import DataARF, DataRMF


class Instrument(xr.Dataset):
    """
    Class to store the data of an instrument
    """

    redistribution: xr.DataArray
    """The photon redistribution probability matrix"""
    area: xr.DataArray
    """The effective area of the instrument"""

    __slots__ = (
        "redistribution",
        "area",
        "e_min_out",
        "e_max_out",
        "e_min_in",
        "e_max_in",
    )

    @classmethod
    def from_matrix(cls, redistribution_matrix, spectral_response, e_min_unfolded, e_max_unfolded, e_min_channel, e_max_channel):
        return cls(
            {
                "redistribution": (
                    ["instrument_channel", "unfolded_channel"],
                    redistribution_matrix,
                    {"description": "Redistribution matrix"},
                ),
                "area": (
                    ["unfolded_channel"],
                    np.array(spectral_response, dtype=np.float64),
                    {"description": "Effective area", "units": "cm^2"},
                ),
            },
            coords={
                "e_min_unfolded": (
                    ["unfolded_channel"],
                    np.array(e_min_unfolded, dtype=np.float64),
                    {"description": "Low bin energy for ingoing channels", "units": "keV"},
                ),
                "e_max_unfolded": (
                    ["unfolded_channel"],
                    np.array(e_max_unfolded, dtype=np.float64),
                    {"description": "High bin energy for ingoing channels", "units": "keV"},
                ),
                "e_min_channel": (
                    ["instrument_channel"],
                    np.array(e_min_channel, dtype=np.float64),
                    {"description": "Low bin energy for outgoing channels", "units": "keV"},
                ),
                "e_max_channel": (
                    ["instrument_channel"],
                    np.array(e_max_channel, dtype=np.float64),
                    {"description": "High bin energy for outgoing channels", "units": "keV"},
                ),
            },
            attrs={"description": "X-ray instrument response dataset"},
        )

    @classmethod
    def from_ogip_file(cls, rmf_path: str | os.PathLike, arf_path: str | os.PathLike = None):
        """
        Load the data from OGIP files.

        Parameters:
            rmf_path: The RMF file path.
            arf_path: The ARF file path.
        """

        rmf = DataRMF.from_file(rmf_path)

        if arf_path is not None:
            specresp = DataARF.from_file(arf_path).specresp

        else:
            specresp = np.ones(rmf.energ_lo.shape)

        return cls.from_matrix(rmf.sparse_matrix, specresp, rmf.energ_lo, rmf.energ_hi, rmf.e_min, rmf.e_max)

    def plot_redistribution(self, **kwargs):
        """
        Plot the redistribution probability matrix

        Parameters:
            **kwargs : `kwargs` passed to https://docs.xarray.dev/en/latest/generated/xarray.plot.pcolormesh.html#xarray.plot.pcolormesh
        """
        import cmasher as cmr

        return xr.plot.pcolormesh(
            self.redistribution,
            x="e_max_unfolded",
            y="e_max_channel",
            xscale="log",
            yscale="log",
            cmap=cmr.ember_r,
            norm=colors.LogNorm(vmin=1e-6, vmax=1),
            add_labels=True,
            **kwargs,
        )

    def plot_area(self, **kwargs):
        """
        Plot the effective area

        Parameters:
            **kwargs : `kwargs` passed to https://docs.xarray.dev/en/latest/generated/xarray.DataArray.plot.line.html#xarray.DataArray.plot.line
        """

        return self.area.plot.step(x="e_min_unfolded", xscale="log", yscale="log", where="post", **kwargs)



from __future__ import annotations

import haiku as hk
import jax.numpy as jnp
import numpy as np

from astropy.table import Table
from haiku.initializers import Constant as HaikuConstant

from ..util.online_storage import table_manager
from .abc import MultiplicativeComponent


class Expfac(MultiplicativeComponent):
    r"""
    An exponential modification of a spectrum.

    $$
    \mathcal{M}(E) = \begin{cases}1 + A \exp \left(-fE\right) &
    \text{if $E>E_c$}\\1 & \text{if $E<E_c$}\end{cases}
    $$

    ??? abstract "Parameters"
        * $A$ : Amplitude of the modification $\left[\text{dimensionless}\right]$
        * $f$ : Exponential factor $\left[\text{keV}^{-1}\right]$
        * $E_c$ : Start energy of modification $\left[\text{keV}\right]$

    """

    def continuum(self, energy):
        amplitude = hk.get_parameter("A", [], float, init=HaikuConstant(1))
        factor = hk.get_parameter("f", [], float, init=HaikuConstant(1))
        pivot = hk.get_parameter("E_c", [], float, init=HaikuConstant(1))

        return jnp.where(energy >= pivot, 1.0 + amplitude * jnp.exp(-factor * energy), 1.0)


class Tbabs(MultiplicativeComponent):
    r"""
    The Tuebingen-Boulder ISM absorption model. This model calculates the cross section for X-ray absorption by the ISM
    as the sum of the cross sections for X-ray absorption due to the gas-phase ISM, the grain-phase ISM,
    and the molecules in the ISM.

    $$
        \mathcal{M}(E) = \exp^{-N_{\text{H}}\sigma(E)}
    $$

    ??? abstract "Parameters"
        * $N_{\text{H}}$ : Equivalent hydrogen column density
            $\left[\frac{\text{atoms}~10^{22}}{\text{cm}^2}\right]$

    !!! note
        Abundances and cross-sections $\sigma$ can be found in Wilms et al. (2000).

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        table = Table.read(table_manager.fetch("xsect_tbabs_wilm.fits"))
        self.energy = jnp.asarray(np.array(table["ENERGY"]), dtype=np.float64)
        self.sigma = jnp.asarray(np.array(table["SIGMA"]), dtype=np.float64)

    def continuum(self, energy):
        nh = hk.get_parameter("N_H", [], float, init=HaikuConstant(1))
        sigma = jnp.interp(energy, self.energy, self.sigma, left=1e9, right=0.0)

        return jnp.exp(-nh * sigma)


class Phabs(MultiplicativeComponent):
    r"""
    A photoelectric absorption model.

    ??? abstract "Parameters"
        * $N_{\text{H}}$ : Equivalent hydrogen column density
            $\left[\frac{\text{atoms}~10^{22}}{\text{cm}^2}\right]$

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        table = Table.read(table_manager.fetch("xsect_phabs_aspl.fits"))
        self.energy = jnp.asarray(np.array(table["ENERGY"]), dtype=np.float64)
        self.sigma = jnp.asarray(np.array(table["SIGMA"]), dtype=np.float64)

    def continuum(self, energy):
        nh = hk.get_parameter("N_H", [], float, init=HaikuConstant(1))
        sigma = jnp.interp(energy, self.energy, self.sigma, left=jnp.inf, right=0.0)

        return jnp.exp(-nh * sigma)


class Wabs(MultiplicativeComponent):
    r"""
    A photo-electric absorption using Wisconsin (Morrison & McCammon 1983) cross-sections.

    ??? abstract "Parameters"
        * $N_{\text{H}}$ : Equivalent hydrogen column density
            $\left[\frac{\text{atoms}~10^{22}}{\text{cm}^2}\right]$

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        table = Table.read(table_manager.fetch("xsect_wabs_angr.fits"))
        self.energy = jnp.asarray(np.array(table["ENERGY"]), dtype=np.float64)
        self.sigma = jnp.asarray(np.array(table["SIGMA"]), dtype=np.float64)

    def continuum(self, energy):
        nh = hk.get_parameter("N_H", [], float, init=HaikuConstant(1))
        sigma = jnp.interp(energy, self.energy, self.sigma, left=1e9, right=0.0)

        return jnp.exp(-nh * sigma)


class Gabs(MultiplicativeComponent):
    r"""
    A Gaussian absorption model.

    $$
        \mathcal{M}(E) = \exp \left( - \frac{\tau}{\sqrt{2 \pi} \sigma} \exp
        \left( -\frac{\left(E-E_0\right)^2}{2 \sigma^2} \right) \right)
    $$

    ??? abstract "Parameters"
        * $\tau$ : Absorption strength $\left[\text{dimensionless}\right]$
        * $\sigma$ : Absorption width $\left[\text{keV}\right]$
        * $E_0$ : Absorption center $\left[\text{keV}\right]$

    !!! note
        The optical depth at line center is $\tau/(\sqrt{2 \pi} \sigma)$.

    """

    def continuum(self, energy):
        tau = hk.get_parameter("tau", [], float, init=HaikuConstant(1))
        sigma = hk.get_parameter("sigma", [], float, init=HaikuConstant(1))
        center = hk.get_parameter("E_0", [], float, init=HaikuConstant(1))

        return jnp.exp(
            -tau / (jnp.sqrt(2 * jnp.pi) * sigma) * jnp.exp(-0.5 * ((energy - center) / sigma) ** 2)
        )


class Highecut(MultiplicativeComponent):
    r"""
    A high-energy cutoff model.

    $$
        \mathcal{M}(E) = \begin{cases} \exp
        \left( \frac{E_c - E}{E_f} \right)& \text{if $E > E_c$}\\ 1 & \text{if $E < E_c$}\end{cases}
    $$

    ??? abstract "Parameters"
        * $E_c$ : Cutoff energy $\left[\text{keV}\right]$
        * $E_f$ : Folding energy $\left[\text{keV}\right]$
    """

    def continuum(self, energy):
        cutoff = hk.get_parameter("E_c", [], float, init=HaikuConstant(1))
        folding = hk.get_parameter("E_f", [], float, init=HaikuConstant(1))

        return jnp.where(energy <= cutoff, 1.0, jnp.exp((cutoff - energy) / folding))


class Zedge(MultiplicativeComponent):
    r"""
    A redshifted absorption edge model.

    $$
        \mathcal{M}(E) = \begin{cases} \exp \left( -D \left(\frac{E(1+z)}{E_c}\right)^3 \right)
        & \text{if $E > E_c$}\\ 1 & \text{if $E < E_c$}\end{cases}
    $$

    ??? abstract "Parameters"
        * $E_c$ : Threshold energy
        * $E_f$ : Absorption depth at the threshold
        * $z$ : Redshift [dimensionless]
    """

    def continuum(self, energy):
        E_c = hk.get_parameter("E_c", [], float, init=HaikuConstant(1))
        D = hk.get_parameter("D", [], float, init=HaikuConstant(1))
        z = hk.get_parameter("z", [], float, init=HaikuConstant(0))

        return jnp.where(energy <= E_c, 1.0, jnp.exp(-D * (energy * (1 + z) / E_c) ** 3))


class Tbpcf(MultiplicativeComponent):
    r"""
    Partial covering model using the Tuebingen-Boulder ISM absorption model (for more details, see `Tbabs`).

    $$
        \mathcal{M}(E) = f \exp^{-N_{\text{H}}\sigma(E)} + (1-f)
    $$

    ??? abstract "Parameters"
        * $N_{\text{H}}$ : Equivalent hydrogen column density
            $\left[\frac{\text{atoms}~10^{22}}{\text{cm}^2}\right]$
        * $f$ : Partial covering fraction, ranges between 0 and 1 [dimensionless]

    !!! note
        Abundances and cross-sections $\sigma$ can be found in Wilms et al. (2000).

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        table = Table.read(table_manager.fetch("xsect_tbabs_wilm.fits"))
        self.energy = jnp.asarray(np.array(table["ENERGY"]), dtype=np.float64)
        self.sigma = jnp.asarray(np.array(table["SIGMA"]), dtype=np.float64)

    def continuum(self, energy):
        nh = hk.get_parameter("N_H", [], float, init=HaikuConstant(1))
        f = hk.get_parameter("f", [], float, init=HaikuConstant(0.2))
        sigma = jnp.interp(energy, self.energy, self.sigma, left=1e9, right=0.0)

        return f * jnp.exp(-nh * sigma) + (1 - f)


class FDcut(MultiplicativeComponent):
    r"""
    A Fermi-Dirac cutoff model.

    $$
        \mathcal{M}(E) = \left( 1 + \exp \left( \frac{E - E_c}{E_f} \right) \right)^{-1}
    $$

    ??? abstract "Parameters"
        * $E_c$ : Cutoff energy $\left[\text{keV}\right]$
        * $E_f$ : Folding energy $\left[\text{keV}\right]$
    """

    def continuum(self, energy):
        cutoff = hk.get_parameter("E_c", [], init=HaikuConstant(1))
        folding = hk.get_parameter("E_f", [], init=HaikuConstant(1))

        return (1 + jnp.exp((energy - cutoff) / folding)) ** -1



from __future__ import annotations

from abc import ABC, abstractmethod
from uuid import uuid4

import haiku as hk
import jax
import jax.numpy as jnp
import networkx as nx
import rich

from haiku._src import base
from jax.scipy.integrate import trapezoid
from rich.table import Table
from simpleeval import simple_eval


class SpectralModel:
    """
    This class is supposed to handle the composition of models through basic
    operations, and allows tracking of the operation graph and individual parameters.
    """

    raw_graph: nx.DiGraph
    graph: nx.DiGraph
    labels: dict[str, str]
    n_parameters: int

    def __init__(self, internal_graph, labels):
        self.raw_graph = internal_graph
        self.labels = labels
        self.graph = self.build_namespace()

        self.n_parameters = hk.data_structures.tree_size(self.params)

    @classmethod
    def from_string(cls, string: str) -> SpectralModel:
        """
        This constructor enable to build a model from a string. The string should be a valid python expression, with
        the following constraints :

        * The model components should be defined in the jaxspec.model.list module
        * The model components should be separated by a * or a + (no convolution yet)
        * The model components should be written with their parameters in parentheses

        Parameters:
            string : The string to parse

        Examples:
            An absorbed model with a powerlaw and a blackbody:

            >>> model = SpectralModel.from_string("Tbabs()*(Powerlaw() + Blackbody())")
        """

        from .list import model_components

        return simple_eval(string, functions=model_components)

    def to_string(self) -> str:
        """
        This method return the string representation of the model.

        Examples:
            Build a model from a string and convert it back to a string:

            >>> model = SpectralModel.from_string("Tbabs()*(Powerlaw() + Blackbody())")
            >>> model.to_string()
            "Tbabs()*(Powerlaw() + Blackbody())"
        """

        return str(self)

    def __str__(self) -> SpectralModel:
        def build_expression(node_id):
            node = self.graph.nodes[node_id]
            if node["type"] == "component":
                string = node["component"].__name__

                if node["kwargs"]:
                    kwargs = ", ".join([f"{k}={v}" for k, v in node["kwargs"].items()])
                    string += f"({kwargs})"
                else:
                    string += "()"
                return string

            elif node["type"] == "operation":
                predecessors = list(self.graph.predecessors(node_id))
                operands = [build_expression(pred) for pred in predecessors]
                operation = node["operation_label"]
                return f"({f' {operation} '.join(operands)})"
            elif node["type"] == "out":
                predecessors = list(self.graph.predecessors(node_id))
                return build_expression(predecessors[0])

        return build_expression("out")[1:-1]

    @property
    def transformed_func_photon(self):
        def func_to_transform(e_low, e_high, n_points=2):
            return self.flux(e_low, e_high, n_points=n_points, energy_flux=False)

        return hk.without_apply_rng(hk.transform(func_to_transform))

    @property
    def transformed_func_energy(self):
        def func_to_transform(e_low, e_high, n_points=2):
            return self.flux(e_low, e_high, n_points=n_points, energy_flux=True)

        return hk.without_apply_rng(hk.transform(func_to_transform))

    @property
    def params(self):
        return self.transformed_func_photon.init(None, jnp.ones(10), jnp.ones(10))

    def __rich_repr__(self):
        table = Table(title=str(self))

        table.add_column("Component", justify="right", style="bold", no_wrap=True)
        table.add_column("Parameter")

        params = self.params

        for component in params.keys():
            once = True

            for parameters in params[component].keys():
                table.add_row(component if once else "", parameters)
                once = False

        return table

    def __repr_html_(self):
        return self.__rich_repr__()

    def __repr__(self):
        if not base.frame_stack:
            rich.print(self.__rich_repr__())
        return ""

    def photon_flux(self, params, e_low, e_high, n_points=2):
        r"""
        Compute the expected counts between $E_\min$ and $E_\max$ by integrating the model.

        $$ \Phi_{\text{photon}}\left(E_\min, ~E_\max\right) =
        \int _{E_\min}^{E_\max}\text{d}E ~ \mathcal{M}\left( E \right)
        \quad \left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$$

        Parameters:
            params : The parameters of the model.
            e_low : The lower bound of the energy bins.
            e_high : The upper bound of the energy bins.
            n_points : The number of points used to integrate the model in each bin.

        !!! info
            This method is internally used in the inference process and should not be used directly. See
            [`photon_flux`][jaxspec.analysis.results.FitResult.photon_flux] to compute
            the photon flux associated with a set of fitted parameters in a
            [`FitResult`][jaxspec.analysis.results.FitResult]
            instead.
        """

        params = jax.tree_map(lambda x: jnp.asarray(x), params)
        e_low = jnp.asarray(e_low)
        e_high = jnp.asarray(e_high)

        return self.transformed_func_photon.apply(params, e_low, e_high, n_points=n_points)

    def energy_flux(self, params, e_low, e_high, n_points=2):
        r"""
        Compute the expected energy flux between $E_\min$ and $E_\max$ by integrating the model.

        $$ \Phi_{\text{energy}}\left(E_\min, ~E_\max\right) =
        \int _{E_\min}^{E_\max}\text{d}E ~ E ~ \mathcal{M}\left( E \right)
        \quad \left[\frac{\text{keV}}{\text{cm}^2\text{s}}\right]$$

        Parameters:
            params : The parameters of the model.
            e_low : The lower bound of the energy bins.
            e_high : The upper bound of the energy bins.
            n_points : The number of points used to integrate the model in each bin.

        !!! info
            This method is internally used in the inference process and should not be used directly. See
            [`energy_flux`](/references/results/#jaxspec.analysis.results.FitResult.energy_flux) to compute
            the energy flux associated with a set of fitted parameters in a
            [`FitResult`](/references/results/#jaxspec.analysis.results.FitResult)
            instead.
        """

        params = jax.tree_map(lambda x: jnp.asarray(x), params)
        e_low = jnp.asarray(e_low)
        e_high = jnp.asarray(e_high)

        return self.transformed_func_energy.apply(params, e_low, e_high, n_points=n_points)

    def build_namespace(self):
        """
        This method build a namespace for the model components, to avoid name collision
        """

        name_space = []
        new_graph = self.raw_graph.copy()

        for node_id in nx.dag.topological_sort(new_graph):
            node = new_graph.nodes[node_id]

            if node and node["type"] == "component":
                name_space.append(node["name"])
                n = name_space.count(node["name"])
                nx.set_node_attributes(new_graph, {node_id: name_space[-1] + f"_{n}"}, "name")

        return new_graph

    def flux(self, e_low, e_high, energy_flux=False, n_points=2):
        """
        This method return the expected counts between e_low and e_high by integrating the model.
        It contains most of the "usine à gaz" which makes jaxspec works.
        It evaluates the graph of operations and returns the result.
        It should be transformed using haiku.
        """

        # TODO : enable interpolation and integration with more than 2 points for the continuum

        if n_points == 2:
            energies = jnp.hstack((e_low, e_high[-1]))
            energies_to_integrate = jnp.stack((e_low, e_high))

        else:
            energies_to_integrate = jnp.linspace(e_low, e_high, n_points)
            energies = energies_to_integrate

        fine_structures_flux = jnp.zeros_like(e_low)
        runtime_modules = {}
        continuum = {}

        # Iterate through the graph in topological order and
        # compute the continuum contribution for each component

        for node_id in nx.dag.topological_sort(self.graph):
            node = self.graph.nodes[node_id]

            # Instantiate the haiku modules
            if node and node["type"] == "component":
                runtime_modules[node_id] = node["component"](name=node["name"], **node["kwargs"])
                continuum[node_id] = runtime_modules[node_id].continuum(energies)

            elif node and node["type"] == "operation":
                component_1 = list(self.graph.in_edges(node_id))[0][0]  # noqa: RUF015
                component_2 = list(self.graph.in_edges(node_id))[1][0]
                continuum[node_id] = node["function"](
                    continuum[component_1], continuum[component_2]
                )

        if n_points == 2:
            flux_1D = continuum[list(self.graph.in_edges("out"))[0][0]]  # noqa: RUF015
            flux = jnp.stack((flux_1D[:-1], flux_1D[1:]))

        else:
            flux = continuum[list(self.graph.in_edges("out"))[0][0]]  # noqa: RUF015

        if energy_flux:
            continuum_flux = trapezoid(
                flux * energies_to_integrate**2,
                x=jnp.log(energies_to_integrate),
                axis=0,
            )

        else:
            continuum_flux = trapezoid(
                flux * energies_to_integrate, x=jnp.log(energies_to_integrate), axis=0
            )

        # Iterate from the root nodes to the output node and
        # compute the fine structure contribution for each component

        root_nodes = [
            node_id
            for node_id, in_degree in self.graph.in_degree(self.graph.nodes)
            if in_degree == 0 and self.graph.nodes[node_id].get("component_type") == "additive"
        ]

        for root_node_id in root_nodes:
            path = nx.shortest_path(self.graph, source=root_node_id, target="out")
            nodes_id_in_path = [node_id for node_id in path]

            flux_from_component, mean_energy = runtime_modules[root_node_id].emission_lines(
                e_low, e_high
            )

            multiplicative_nodes = []

            # Search all multiplicative components connected to this node
            # and apply them at mean energy
            for node_id in nodes_id_in_path[::-1]:
                multiplicative_nodes.extend(
                    [node_id for node_id in self.find_multiplicative_components(node_id)]
                )

            for mul_node in multiplicative_nodes:
                flux_from_component *= runtime_modules[mul_node].continuum(mean_energy)

            if energy_flux:
                fine_structures_flux += trapezoid(
                    flux_from_component * energies_to_integrate,
                    x=jnp.log(energies_to_integrate),
                    axis=0,
                )

            else:
                fine_structures_flux += flux_from_component

        return continuum_flux + fine_structures_flux

    def find_multiplicative_components(self, node_id):
        """
        Recursively finds all the multiplicative components connected to the node with the given ID.
        """
        node = self.graph.nodes[node_id]
        multiplicative_nodes = []

        if node.get("operation_type") == "mul":
            # Recursively find all the multiplicative components using the predecessors
            predecessors = self.graph.pred[node_id]
            for node_id in predecessors:
                if self.graph.nodes[node_id].get("component_type") == "multiplicative":
                    multiplicative_nodes.append(node_id)
                elif self.graph.nodes[node_id].get("operation_type") == "mul":
                    multiplicative_nodes.extend(self.find_multiplicative_components(node_id))

        return multiplicative_nodes

    def __call__(self, pars, e_low, e_high, **kwargs):
        return self.photon_flux(pars, e_low, e_high, **kwargs)

    @classmethod
    def from_component(cls, component, **kwargs) -> SpectralModel:
        """
        Build a model from a single component
        """

        graph = nx.DiGraph()

        # Add the component node
        # Random static node id to keep it trackable in the graph
        node_id = str(uuid4())

        if component.type == "additive":

            def lam_func(e):
                return (
                    component(**kwargs).continuum(e)
                    + component(**kwargs).emission_lines(e, e + 1)[0]
                )

        elif component.type == "multiplicative":

            def lam_func(e):
                return component().continuum(e)

        else:

            def lam_func(e):
                return print("Some components are not working at this stage")

        node_properties = {
            "type": "component",
            "component_type": component.type,
            "name": component.__name__.lower(),
            "component": component,
            # "params": hk.transform(lam_func).init(None, jnp.ones(1)),
            "fine_structure": False,
            "kwargs": kwargs,
            "depth": 0,
        }

        graph.add_node(node_id, **node_properties)

        # Add the output node
        labels = {node_id: component.__name__.lower(), "out": "out"}

        graph.add_node("out", type="out", depth=1)
        graph.add_edge(node_id, "out")

        return cls(graph, labels)

    def compose(
        self, other: SpectralModel, operation=None, function=None, name=None
    ) -> SpectralModel:
        """
        This function operate a composition between the operation graph of two models
        1) It fuses the two graphs using which joins at the 'out' nodes
        2) It relabels the 'out' node with a unique identifier and labels it with the operation
        3) It links the operation to a new 'out' node
        """

        # Compose the two graphs with their output as common node
        # and add the operation node by overwriting the 'out' node
        node_id = str(uuid4())
        graph = nx.relabel_nodes(nx.compose(self.raw_graph, other.raw_graph), {"out": node_id})
        nx.set_node_attributes(graph, {node_id: "operation"}, "type")
        nx.set_node_attributes(graph, {node_id: operation}, "operation_type")
        nx.set_node_attributes(graph, {node_id: function}, "function")
        nx.set_node_attributes(graph, {node_id: name}, "operation_label")

        # Merge label dictionaries
        labels = self.labels | other.labels
        labels[node_id] = operation

        # Now add the output node and link it to the operation node
        graph.add_node("out", type="out")
        graph.add_edge(node_id, "out")

        # Compute the new depth of each node
        longest_path = nx.dag_longest_path_length(graph)

        for node in graph.nodes:
            nx.set_node_attributes(
                graph,
                {node: longest_path - nx.shortest_path_length(graph, node, "out")},
                "depth",
            )

        return SpectralModel(graph, labels)

    def __add__(self, other: SpectralModel) -> SpectralModel:
        return self.compose(other, operation="add", function=lambda x, y: x + y, name="+")

    def __mul__(self, other: SpectralModel) -> SpectralModel:
        return self.compose(other, operation="mul", function=lambda x, y: x * y, name=r"*")

    def export_to_mermaid(self, file=None):
        mermaid_code = "graph LR\n"  # LR = left to right

        # Add nodes
        for node, attributes in self.graph.nodes(data=True):
            if attributes["type"] == "component":
                name, number = attributes["name"].split("_")
                mermaid_code += f'    {node}("{name.capitalize()} ({number})")\n'

            if attributes["type"] == "operation":
                if attributes["operation_type"] == "add":
                    mermaid_code += f"    {node}{{+}}\n"

                if attributes["operation_type"] == "mul":
                    mermaid_code += f"    {node}{{x}}\n"

            if attributes["type"] == "out":
                mermaid_code += f'    {node}("Output")\n'

        # Draw connexion between nodes
        for source, target in self.graph.edges():
            mermaid_code += f"    {source} --> {target}\n"

        if file is None:
            return mermaid_code
        else:
            with open(file, "w") as f:
                f.write(mermaid_code)

    def plot(self, figsize=(8, 8)):
        import matplotlib.pyplot as plt

        plt.figure(figsize=figsize)

        pos = nx.multipartite_layout(self.graph, subset_key="depth", scale=1)

        nodes_out = [x for x, y in self.graph.nodes(data=True) if y["type"] == "out"]
        nx.draw_networkx_nodes(self.graph, pos, nodelist=nodes_out, node_color="tab:green")
        nx.draw_networkx_edges(self.graph, pos, width=1.0)

        nx.draw_networkx_labels(
            self.graph,
            pos,
            labels=nx.get_node_attributes(self.graph, "name"),
            font_size=12,
            font_color="black",
            bbox={"fc": "tab:red", "boxstyle": "round", "pad": 0.3},
        )
        nx.draw_networkx_labels(
            self.graph,
            pos,
            labels=nx.get_node_attributes(self.graph, "operation_label"),
            font_size=12,
            font_color="black",
            bbox={"fc": "tab:blue", "boxstyle": "circle", "pad": 0.3},
        )

        plt.axis("equal")
        plt.axis("off")
        plt.tight_layout()
        plt.show()


class ComponentMetaClass(type(hk.Module)):
    """
    This metaclass enable the construction of model from components with a simple
    syntax while style enabling the components to be used as haiku modules.
    """

    def __call__(self, **kwargs) -> SpectralModel:
        """
        This method enable to use model components as haiku modules when folded in a haiku transform
        function and also to instantiate them as SpectralModel when out of a haiku transform
        """

        if not base.frame_stack:
            return SpectralModel.from_component(self, **kwargs)

        else:
            return super().__call__(**kwargs)


class ModelComponent(hk.Module, ABC, metaclass=ComponentMetaClass):
    """
    Abstract class for model components
    """

    type: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AdditiveComponent(ModelComponent, ABC):
    type = "additive"

    def continuum(self, energy):
        """
        Method for computing the continuum associated to the model.
        By default, this is set to 0, which means that the model has no continuum.
        This should be overloaded by the user if the model has a continuum.
        """

        return jnp.zeros_like(energy)

    def emission_lines(self, e_min, e_max) -> (jax.Array, jax.Array):
        """
        Method for computing the fine structure of an additive model between two energies.
        By default, this is set to 0, which means that the model has no emission lines.
        This should be overloaded by the user if the model has a fine structure.
        """

        return jnp.zeros_like(e_min), (e_min + e_max) / 2

    '''
    def integral(self, e_min, e_max):
        r"""
        Method for integrating an additive model between two energies. It relies on
        double exponential quadrature for finite intervals to compute an approximation
        of the integral of a model.

        references
        ----------
        * $Takahasi and Mori (1974) <https://ems.press/journals/prims/articles/2686>$_
        * $Mori and Sugihara (2001) <https://doi.org/10.1016/S0377-0427(00)00501-X>$_
        * $Tanh-sinh quadrature <https://en.wikipedia.org/wiki/Tanh-sinh_quadrature>$_ from Wikipedia

        """

        t = jnp.linspace(-4, 4, 71) # The number of points used is hardcoded and this is not ideal
        # Quadrature nodes as defined in reference
        phi = jnp.tanh(jnp.pi / 2 * jnp.sinh(t))
        dphi = jnp.pi / 2 * jnp.cosh(t) * (1 / jnp.cosh(jnp.pi / 2 * jnp.sinh(t)) ** 2)
        # Change of variable to turn the integral from E_min to E_max into an integral from -1 to 1
        x = (e_max - e_min) / 2 * phi + (e_max + e_min) / 2
        dx = (e_max - e_min) / 2 * dphi

        return jnp.trapz(self(x) * dx, x=t)
    '''


class MultiplicativeComponent(ModelComponent, ABC):
    type = "multiplicative"

    @abstractmethod
    def continuum(self, energy): ...



from __future__ import annotations

import astropy.constants
import astropy.units as u
import haiku as hk
import interpax
import jax
import jax.numpy as jnp
import jax.scipy as jsp
import numpy as np

from astropy.table import Table
from haiku.initializers import Constant as HaikuConstant

from ..util.integrate import integrate_interval
from ..util.online_storage import table_manager
from .abc import AdditiveComponent


class Powerlaw(AdditiveComponent):
    r"""
    A power law model

    $$\mathcal{M}\left( E \right) = K \left( \frac{E}{E_0} \right)^{-\alpha}$$

    ??? abstract "Parameters"
        * $\alpha$ : Photon index of the power law $\left[\text{dimensionless}\right]$
        * $E_0$ : Reference energy fixed at 1 keV $\left[ \mathrm{keV}\right]$
        * $K$ : Normalization at the reference energy (1 keV) $\left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$
    """

    def continuum(self, energy):
        alpha = hk.get_parameter("alpha", [], float, init=HaikuConstant(1.3))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1e-4))

        return norm * energy ** (-alpha)


class Additiveconstant(AdditiveComponent):
    r"""
    A constant model

    $$\mathcal{M}\left( E \right) = K$$

    ??? abstract "Parameters"
        * $K$ : Normalization $\left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$
    """

    def continuum(self, energy):
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))

        return norm * jnp.ones_like(energy)

    def primitive(self, energy):
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))

        return norm * energy


class Lorentz(AdditiveComponent):
    r"""
    A Lorentzian line profile

    $$\mathcal{M}\left( E \right) = K\frac{\frac{\sigma}{2\pi}}{(E-E_L)^2 + \left(\frac{\sigma}{2}\right)^2}$$

    ??? abstract "Parameters"
        - $E_L$ : Energy of the line $\left[\text{keV}\right]$
        - $\sigma$ : FWHM of the line $\left[\text{keV}\right]$
        - $K$ : Normalization $\left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$
    """

    def continuum(self, energy):
        line_energy = hk.get_parameter("E_l", [], float, init=HaikuConstant(1))
        sigma = hk.get_parameter("sigma", [], float, init=HaikuConstant(1))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))

        return norm * sigma / (2 * jnp.pi) / ((energy - line_energy) ** 2 + (sigma / 2) ** 2)


class Logparabola(AdditiveComponent):
    r"""
    A LogParabola model

    $$
    \mathcal{M}\left( E \right) = K  \left( \frac{E}{E_{\text{Pivot}}} \right)
    ^{-(\alpha + \beta ~ \log (E/E_{\text{Pivot}})) }
    $$

    ??? abstract "Parameters"
        * $a$ : Slope of the LogParabola at the pivot energy $\left[\text{dimensionless}\right]$
        * $b$ : Curve parameter of the LogParabola $\left[\text{dimensionless}\right]$
        * $E_{\text{Pivot}}$ : Pivot energy fixed at 1 keV $\left[ \mathrm{keV}\right]$
        * $K$ : Normalization at the pivot energy (1keV) $\left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$
    """

    # TODO : conform with xspec definition
    def continuum(self, energy):
        a = hk.get_parameter("a", [], float, init=HaikuConstant(11 / 3))
        b = hk.get_parameter("b", [], float, init=HaikuConstant(0.2))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))

        return norm * energy ** (-(a + b * jnp.log(energy)))


class Blackbody(AdditiveComponent):
    r"""
    A black body model

    $$\mathcal{M}\left( E \right) = \frac{K \times 8.0525 E^{2}}{(k_B T)^{4}\left(\exp(E/k_BT)-1\right)}$$

    ??? abstract "Parameters"
        * $k_B T$ : Temperature $\left[\text{keV}\right]$
        * $K$ : $L_{39}/D_{10}^{2}$, where $L_{39}$ is the source luminosity in units of $10^{39}$ erg/s
        and $D_{10}$ is the distance to the source in units of 10 kpc
    """

    def continuum(self, energy):
        kT = hk.get_parameter("kT", [], float, init=HaikuConstant(11 / 3))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))

        return norm * 8.0525 * energy**2 / ((kT**4) * (jnp.exp(energy / kT) - 1))


class Blackbodyrad(AdditiveComponent):
    r"""
    A black body model in radius normalization

    $$\mathcal{M}\left( E \right) = \frac{K \times 1.0344\times 10^{-3} E^{2}}{\left(\exp (E/k_BT)-1\right)}$$

    ??? abstract "Parameters"
        * $k_B T$ : Temperature $\left[\text{keV}\right]$
        * $K$ : $R^2_{km}/D_{10}^{2}$, where $R_{km}$ is the source radius in km
        and $D_{10}$ is the distance to the source in units of 10 kpc [dimensionless]
    """

    def continuum(self, energy):
        kT = hk.get_parameter("kT", [], float, init=HaikuConstant(11 / 3))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))

        return norm * 1.0344e-3 * energy**2 / (jnp.exp(energy / kT) - 1)


class Gauss(AdditiveComponent):
    r"""
    A Gaussian line profile. If the width is $$\leq 0$$ then it is treated as a delta function.
    The `Zgauss` variant computes a redshifted Gaussian.

    $$\mathcal{M}\left( E \right) = \frac{K}{\sigma \sqrt{2 \pi}}\exp\left(\frac{-(E-E_L)^2}{2\sigma^2}\right)$$

    ??? abstract "Parameters"
        * $E_L$ : Energy of the line $\left[\text{keV}\right]$
        * $\sigma$ : Width of the line $\left[\text{keV}\right]$
        * $K$ : Normalization $\left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$
    """

    def continuum(self, energy) -> (jax.Array, jax.Array):
        line_energy = hk.get_parameter("E_l", [], float, init=HaikuConstant(1))
        sigma = hk.get_parameter("sigma", [], float, init=HaikuConstant(1))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))

        return norm * jsp.stats.norm.pdf(energy, loc=line_energy, scale=sigma)


"""
class APEC(AdditiveComponent):
    def __init__(self, name="apec"):
        super(APEC, self).__init__(name=name)

        ref = importlib.resources.files("jaxspec") / "tables/apec_tab.npz"
        with importlib.resources.as_file(ref) as path:
            files = np.load(path)

        self.kT_ref = files["kT_ref"]
        self.e_ref = np.nan_to_num(files["continuum_energy"], nan=1e6)
        self.c_ref = np.nan_to_num(files["continuum_emissivity"])
        self.pe_ref = np.nan_to_num(files["pseudo_energy"], nan=1e6)
        self.pc_ref = np.nan_to_num(files["pseudo_emissivity"])
        self.energy_lines = np.nan_to_num(files["lines_energy"], nan=1e6)  # .astype(np.float32))
        self.epsilon_lines = np.nan_to_num(files["lines_emissivity"])  # .astype(np.float32))
        self.element_lines = np.nan_to_num(files["lines_element"])  # .astype(np.int32))

        del files

        self.trace_elements = jnp.array([3, 4, 5, 9, 11, 15, 17, 19, 21, 22, 23, 24, 25, 27, 29, 30]) - 1
        self.metals = np.array([6, 7, 8, 10, 12, 13, 14, 16, 18, 20, 26, 28]) - 1  # Element number to python index
        self.metals_one_hot = np.zeros((30,))
        self.metals_one_hot[self.metals] = 1

    def interp_on_cubes(self, energy, energy_cube, continuum_cube):
        # Changer en loginterp
        # Interpoler juste sur les points qui ne sont pas tabulés
        # Ajouter l'info de la taille max des données (à resortir dans la routine qui trie les fichier apec)
        return jnp.vectorize(
            lambda ecube, ccube: jnp.interp(energy, ecube, ccube),
            signature="(k),(k)->()",
        )(energy_cube, continuum_cube)

    def reduction_with_elements(self, Z, energy, energy_cube, continuum_cube):
        return jnp.sum(
            self.interp_on_cubes(energy, energy_cube, continuum_cube) * jnp.where(self.metals_one_hot, Z, 1.0)[None, :],
            axis=-1,
        )

    def mono_fine_structure(self, e_low, e_high) -> (jax.Array, jax.Array):
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))
        kT = hk.get_parameter("kT", [], float, init=HaikuConstant(1))
        Z = hk.get_parameter("Z", [], float, init=HaikuConstant(1))

        idx = jnp.searchsorted(self.kT_ref, kT, side="left") - 1

        energy = jax_slice(self.energy_lines, idx, 2)
        epsilon = jax_slice(self.epsilon_lines, idx, 2)
        element = jax_slice(self.element_lines, idx, 2) - 1

        emissivity_in_bins = jnp.where((e_low < energy) & (energy < e_high), True, False) * epsilon
        flux_at_edges = jnp.nansum(
            jnp.where(jnp.isin(element, self.metals), Z, 1) * emissivity_in_bins,
            axis=-1,
        )  # Coeff for metallicity

        return (
            jnp.interp(kT, jax_slice(self.kT_ref, idx, 2), flux_at_edges) * 1e14 * norm,
            (e_low + e_high) / 2,
        )

    def emission_lines(self, e_low, e_high) -> (jax.Array, jax.Array):
        # Compute the fine structure lines with e_low and e_high as array, mapping the mono_fine_structure function
        # over the various axes of e_low and e_high

        return jnp.vectorize(self.mono_fine_structure)(e_low, e_high)

        # return jnp.zeros_like(e_low), (e_low + e_high)/2

    @partial(jnp.vectorize, excluded=(0,))
    def continuum(self, energy):
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))
        kT = hk.get_parameter("kT", [], float, init=HaikuConstant(1))
        Z = hk.get_parameter("Z", [], float, init=HaikuConstant(1))

        idx = jnp.searchsorted(self.kT_ref, kT, side="left") - 1  # index of left value

        continuum = jnp.interp(
            kT,
            jax_slice(self.kT_ref, idx, 2),
            self.reduction_with_elements(Z, energy, jax_slice(self.e_ref, idx, 2), jax_slice(self.c_ref, idx, 2)),
        )
        pseudo = jnp.interp(
            kT,
            jax_slice(self.kT_ref, idx, 2),
            self.reduction_with_elements(
                Z,
                energy,
                jax_slice(self.pe_ref, idx, 2),
                jax_slice(self.pc_ref, idx, 2),
            ),
        )

        return (continuum + pseudo) * 1e14 * norm
"""


class Cutoffpl(AdditiveComponent):
    r"""
    A power law model with high energy exponential cutoff

    $$\mathcal{M}\left( E \right) = K \left( \frac{E}{E_0} \right)^{-\alpha} \exp(-E/\beta)$$

    ??? abstract "Parameters"
        * $\alpha$ : Photon index of the power law $\left[\text{dimensionless}\right]$
        * $\beta$ : Folding energy of the exponential cutoff $\left[\text{keV}\right]$
        * $E_0$ : Reference energy fixed at 1 keV $\left[ \mathrm{keV}\right]$
        * $K$ : Normalization at the reference energy (1 keV) $\left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$
    """

    def continuum(self, energy):
        alpha = hk.get_parameter("alpha", [], float, init=HaikuConstant(1.3))
        beta = hk.get_parameter("beta", [], float, init=HaikuConstant(15))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1e-4))

        return norm * energy ** (-alpha) * jnp.exp(-energy / beta)


'''
class Diskpbb(AdditiveComponent):
    r"""
    A multiple blackbody disk model where local disk temperature T(r) is proportional to $$r^{(-p)}$$,
    where $$p$$ is a free parameter. The standard disk model, diskbb, is recovered if $$p=0.75$$.
    If radial advection is important then $$p<0.75$$.

    $$\\mathcal{M}\\left( E \right) = \frac{2\\pi(\\cos i)r^{2}_{\text{in}}}{pd^2} \\int_{T_{\text{in}}}^{T_{\text{out}}}
    \\left( \frac{T}{T_{\text{in}}} \right)^{-2/p-1} \text{bbody}(E,T) \frac{dT}{T_{\text{in}}}$$

    ??? abstract "Parameters"
        * $\text{norm}$ : $\\cos i(r_{\text{in}}/d)^{2}$,
        where $r_{\text{in}}$ is "an apparent" inner disk radius $\\left[\text{km}\right]$,
        $d$ the distance to the source in units of $10 \text{kpc}$,
        $i$ the angle of the disk ($i=0$ is face-on)
        * $p$ : Exponent of the radial dependence of the disk temperature $\\left[\text{dimensionless}\right]$
        * $T_{\text{in}}$ : Temperature at inner disk radius $\\left[ \\mathrm{keV}\right]$
    """

    def continuum(self, energy):
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))
        p = hk.get_parameter("p", [], float, init=HaikuConstant(0.75))
        tin = hk.get_parameter("Tin", [], float, init=HaikuConstant(1))

        # Tout is set to 0 as it is evaluated at R=infinity
        def integrand(kT, energy):
            return 2.78e-3 * energy**2 * (kT / tin) ** (-2 / p - 1) / (jnp.exp(energy / kT) - 1)

        func_vmapped = jax.vmap(lambda e: integrate_interval(lambda kT: integrand(kT, e), 0, tin, n=51))

        return norm * (0.75 / p) * func_vmapped(energy)
'''


class Diskbb(AdditiveComponent):
    r"""
    `Diskpbb` with $p=0.75$

    ??? abstract "Parameters"
        * $T_{\text{in}}$ : Temperature at inner disk radius $\left[ \mathrm{keV}\right]$
        * $\text{norm}$ : $\cos i(r_{\text{in}}/d)^{2}$,
        where $r_{\text{in}}$ is "an apparent" inner disk radius $\left[\text{km}\right]$,
        $d$ the distance to the source in units of $10 \text{kpc}$, $i$ the angle of the disk ($i=0$ is face-on)
    """

    def continuum(self, energy):
        p = 0.75
        tout = 0.0
        tin = hk.get_parameter("Tin", [], float, init=HaikuConstant(1))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))

        # Tout is set to 0 as it is evaluated at R=infinity
        def integrand(kT, e, tin, p):
            return e**2 * (kT / tin) ** (-2 / p - 1) / (jnp.exp(e / kT) - 1)

        integral = integrate_interval(integrand)
        return (
            norm
            * 2.78e-3
            * (0.75 / p)
            / tin
            * jnp.vectorize(lambda e: integral(tout, tin, e, tin, p))(energy)
        )


class Agauss(AdditiveComponent):
    r"""
    A simple Gaussian line profile in wavelength space.
    If the width is $\leq 0$ then it is treated as a delta function.
    The `Zagauss` variant computes a redshifted Gaussian.

    $$\mathcal{M}\left( \lambda \right) =
    \frac{K}{\sigma \sqrt{2 \pi}} \exp\left(\frac{-(\lambda - \lambda_L)^2}{2 \sigma^2}\right)$$

    ??? abstract "Parameters"
        * $\lambda_L$ : Wavelength of the line in Angström $\left[\unicode{x212B}\right]$
        * $\sigma$ : Width of the line width in Angström  $\left[\unicode{x212B}\right]$
        * $K$ : Normalization $\left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$
    """

    def continuum(self, energy) -> (jax.Array, jax.Array):
        hc = (astropy.constants.h * astropy.constants.c).to(u.angstrom * u.keV).value
        line_wavelength = hk.get_parameter("Lambda_l", [], float, init=HaikuConstant(hc))
        sigma = hk.get_parameter("sigma", [], float, init=HaikuConstant(0.001))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))

        return norm * jsp.stats.norm.pdf(hc / energy, loc=line_wavelength, scale=sigma)


class Zagauss(AdditiveComponent):
    r"""
    A redshifted Gaussian line profile in wavelength space.
    If the width is $\leq 0$ then it is treated as a delta function.

    $$\mathcal{M}\left( \lambda \right) =
    \frac{K (1+z)}{\sigma \sqrt{2 \pi}} \exp\left(\frac{-(\lambda/(1+z) - \lambda_L)^2}{2 \sigma^2}\right)$$

    ??? abstract "Parameters"
        * $\lambda_L$ : Wavelength of the line in Angström $\left[\text{\AA}\right]$
        * $\sigma$ : Width of the line width in Angström $\left[\text{\AA}\right]$
        * $K$ : Normalization $\left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$
        * $z$ : Redshift [dimensionless]
    """

    def continuum(self, energy) -> (jax.Array, jax.Array):
        hc = (astropy.constants.h * astropy.constants.c).to(u.angstrom * u.keV).value
        line_wavelength = hk.get_parameter("Lambda_l", [], float, init=HaikuConstant(hc))
        sigma = hk.get_parameter("sigma", [], float, init=HaikuConstant(0.001))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))
        redshift = hk.get_parameter("redshift", [], float, init=HaikuConstant(0))

        return (
            norm
            * (1 + redshift)
            * jsp.stats.norm.pdf((hc / energy) / (1 + redshift), loc=line_wavelength, scale=sigma)
        )


class Zgauss(AdditiveComponent):
    r"""
    A redshifted Gaussian line profile. If the width is $\leq 0$ then it is treated as a delta function.

    $$\mathcal{M}\left( E \right) =
    \frac{K}{(1+z) \sigma \sqrt{2 \pi}}\exp\left(\frac{-(E(1+z)-E_L)^2}{2\sigma^2}\right)$$

    ??? abstract "Parameters"
        * $E_L$ : Energy of the line $\left[\text{keV}\right]$
        * $\sigma$ : Width of the line $\left[\text{keV}\right]$
        * $K$ : Normalization $\left[\frac{\text{photons}}{\text{cm}^2\text{s}}\right]$
        * $z$ : Redshift [dimensionless]
    """

    def continuum(self, energy) -> (jax.Array, jax.Array):
        line_energy = hk.get_parameter("E_l", [], float, init=HaikuConstant(1))
        sigma = hk.get_parameter("sigma", [], float, init=HaikuConstant(1))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1))
        redshift = hk.get_parameter("redshift", [], float, init=HaikuConstant(0))

        return (norm / (1 + redshift)) * jsp.stats.norm.pdf(
            energy * (1 + redshift), loc=line_energy, scale=sigma
        )


class NSatmos(AdditiveComponent):
    r"""
    A neutron star atmosphere model based on the `NSATMOS` model from `XSPEC`. See [this page](https://heasarc.gsfc.nasa.gov/docs/xanadu/xspec/manual/node205.html)

    !!! warning
        The boundary case of $R_{\text{NS}} < 1.125 R_{\text{S}}$ is handled with a null flux instead of a constant value as in `XSPEC`.

    ??? abstract "Parameters"
        * $T_{eff}$ : Effective temperature at the surface in K (No redshift applied)
        * $M_{ns}$ : Mass of the NS in solar masses
        * $R_∞$ : Radius at infinity (modulated by gravitational effects) in km
        * $D$ : Distance to the neutron star in kpc
        * norm : fraction of the neutron star surface emitting
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        entry_table = Table.read(table_manager.fetch("nsatmosdata.fits"), 1)

        # Extract the table values. All this code could be summarize in two lines if we reformat the nsatmosdata.fits table
        self.tab_temperature = np.asarray(entry_table["TEMP"][0], dtype=float)  # Logarithmic value
        self.tab_gravity = np.asarray(entry_table["GRAVITY"][0], dtype=float)  # Logarithmic value
        self.tab_mucrit = np.asarray(entry_table["MUCRIT"][0], dtype=float)
        self.tab_energy = np.asarray(entry_table["ENERGY"][0], dtype=float)
        self.tab_flux_flat = Table.read(table_manager.fetch("nsatmosdata.fits"), 2)["FLUX"]

        tab_flux = np.empty(
            (
                self.tab_temperature.size,
                self.tab_gravity.size,
                self.tab_mucrit.size,
                self.tab_energy.size,
            )
        )

        for i in range(len(self.tab_temperature)):
            for j in range(len(self.tab_gravity)):
                for k in range(len(self.tab_mucrit)):
                    tab_flux[i, j, k] = np.array(
                        self.tab_flux_flat[
                            i * len(self.tab_gravity) * len(self.tab_mucrit)
                            + j * len(self.tab_mucrit)
                            + k
                        ]
                    )

        self.tab_flux = np.asarray(tab_flux, dtype=float)

    def interp_flux_func(self, temperature_log, gravity_log, mu):
        # Interpolate the tables to get the flux on the tabulated energy grid

        return interpax.interp3d(
            10.0**temperature_log,
            10.0**gravity_log,
            mu,
            10.0**self.tab_temperature,
            10.0**self.tab_gravity,
            self.tab_mucrit,
            self.tab_flux,
            method="linear",
        )

    def continuum(self, energy):
        temp_log = hk.get_parameter(
            "Tinf", [], float, init=HaikuConstant(6.0)
        )  # log10 of temperature in Kelvin

        #         'Tinf': temp_log, # 5 to 6.5
        #         'M': mass, # 0.5 to 3
        #         'Rns': radius, # 5 to 30

        mass = hk.get_parameter("M", [], float, init=HaikuConstant(1.4))
        radius = hk.get_parameter("Rns", [], float, init=HaikuConstant(10.0))
        distance = hk.get_parameter("dns", [], float, init=HaikuConstant(10.0))
        norm = hk.get_parameter("norm", [], float, init=HaikuConstant(1.0))

        # Derive parameters usable to retrive value in the flux table
        Rcgs = 1e5 * radius  # Radius in cgs
        r_schwarzschild = 2.95e5 * mass  # Schwarzschild radius in cgs
        r_normalized = Rcgs / r_schwarzschild  # Ratio of the radius to the Schwarzschild radius
        r_over_Dsql = 2 * jnp.log10(
            Rcgs / (3.09e21 * distance)
        )  # Log( (R/D)**2 ), 3.09e21 constant transforms radius in cgs to kpc
        zred1 = 1 / jnp.sqrt(1 - (1 / r_normalized))  # Gravitational redshift
        gravity = (6.67e-8 * 1.99e33 * mass / Rcgs**2) * zred1  # Gravity field g in cgs
        gravity_log = jnp.log10(
            gravity
        )  # Log gravity because this is the format given in the table

        # Not sure about mu yet, but it is linked to causality
        cmu = jnp.where(
            r_normalized < 1.5, jnp.sqrt(1.0 - 6.75 / r_normalized**2 + 6.75 / r_normalized**3), 0.0
        )

        # Interpolate the flux table to get the flux at the surface

        flux = jax.jit(self.interp_flux_func)(temp_log, gravity_log, cmu)

        # Rescale the photon energies and fluxes back to the correct local temperature
        Tfactor = 10.0 ** (temp_log - 6.0)
        fluxshift = 3.0 * (temp_log - 6.0)
        E = self.tab_energy * Tfactor
        flux += fluxshift

        # Rescale applying redshift
        fluxshift = -jnp.log10(zred1)
        E = E / zred1
        flux += fluxshift

        # Convert to counts/keV (which corresponds to dividing by1.602e-9*EkeV)
        # Multiply by the area of the star, and calculate the count rate at the observer
        flux += r_over_Dsql
        counts = 10.0 ** (flux - jnp.log10(1.602e-9) - jnp.log10(E))

        true_flux = norm * jnp.exp(
            interpax.interp1d(jnp.log(energy), jnp.log(E), jnp.log(counts), method="linear")
        )

        return jax.lax.select(r_normalized < 1.125, jnp.zeros_like(true_flux), true_flux)



from collections.abc import Callable
from contextlib import contextmanager
from time import perf_counter

import haiku as hk

from jax.random import PRNGKey, split


@contextmanager
def catchtime(desc="Task", print_time=True) -> Callable[[], float]:
    """
    Context manager to measure time taken by a task.

    Parameters
    ----------
        desc (str): Description of the task.
        print_time (bool): Whether to print the time taken by the task.

    Returns
    -------
        Callable[[], float]: Function to get the time taken by the task.
    """

    t1 = t2 = perf_counter()
    yield lambda: t2 - t1
    t2 = perf_counter()
    if print_time:
        print(f"{desc}: {t2 - t1:.3f} seconds")


def sample_prior(dict_of_prior, key=PRNGKey(0), flat_parameters=False):
    """
    Sample the prior distribution from a dict of distributions
    """

    parameters = dict(hk.data_structures.to_haiku_dict(dict_of_prior))
    parameters_flat = {}

    for m, n, distribution in hk.data_structures.traverse(dict_of_prior):
        key, subkey = split(key)
        parameters[m][n] = distribution.sample(subkey)
        parameters_flat[m + "_" + n] = distribution.sample(subkey)

    return parameters if not flat_parameters else parameters_flat



import jax
import numpyro
import haiku as hk
import numpy as np
import jax.numpy as jnp
from typing import Callable
from jax.experimental.sparse import BCOO
from typing import TYPE_CHECKING
from numpyro.distributions import Poisson
from jax.typing import ArrayLike
from numpyro.distributions import Distribution


if TYPE_CHECKING:
    from ..model.abc import SpectralModel
    from ..data import ObsConfiguration
    from ..util.typing import PriorDictModel, PriorDictType



class CountForwardModel(hk.Module):
    """
    A haiku module which allows to build the function that simulates the measured counts
    """

    # TODO: It has no point of being a haiku module, it should be a simple function

    def __init__(self, model: 'SpectralModel', folding: 'ObsConfiguration', sparse=False):
        super().__init__()
        self.model = model
        self.energies = jnp.asarray(folding.in_energies)

        if (
            sparse
        ):  # folding.transfer_matrix.data.density > 0.015 is a good criterion to consider sparsify
            self.transfer_matrix = BCOO.from_scipy_sparse(
                folding.transfer_matrix.data.to_scipy_sparse().tocsr()
            )

        else:
            self.transfer_matrix = jnp.asarray(folding.transfer_matrix.data.todense())

    def __call__(self, parameters):
        """
        Compute the count functions for a given observation.
        """

        expected_counts = self.transfer_matrix @ self.model.photon_flux(parameters, *self.energies)

        return jnp.clip(expected_counts, a_min=1e-6)


def forward_model(
        model: 'SpectralModel',
        parameters,
        obs_configuration: 'ObsConfiguration',
        sparse=False,
    ):

    energies = np.asarray(obs_configuration.in_energies)

    if sparse:
        # folding.transfer_matrix.data.density > 0.015 is a good criterion to consider sparsify
        transfer_matrix = BCOO.from_scipy_sparse(
            obs_configuration.transfer_matrix.data.to_scipy_sparse().tocsr()
        )

    else:
        transfer_matrix = np.asarray(obs_configuration.transfer_matrix.data.todense())

    expected_counts = transfer_matrix @ model.photon_flux(parameters, *energies)

    # The result is clipped at 1e-6 to avoid 0 round-off and diverging likelihoods
    return jnp.clip(expected_counts, a_min=1e-6)


def build_numpyro_model_for_single_obs(
    obs,
    model,
    background_model,
    name: str = "",
    sparse: bool = False,
) -> Callable:
    """
    Build a numpyro model for a given observation and spectral model.
    """

    def numpyro_model(prior_params, observed=True):

        # Return the expected countrate for a set of parameters
        obs_model = jax.jit(lambda par: forward_model(model, par, obs, sparse=sparse))
        countrate = obs_model(prior_params)

        # Handle the background model
        if (getattr(obs, "folded_background", None) is not None) and (background_model is not None):
            bkg_countrate = background_model.numpyro_model(
                obs, model, name="bkg_" + name, observed=observed
            )

        elif (getattr(obs, "folded_background", None) is None) and (background_model is not None):
            raise ValueError(
                "Trying to fit a background model but no background is linked to this observation"
            )

        else:
            bkg_countrate = 0.0


        # Register the observed value
        # This is the case where we fit a model to a TOTAL spectrum as defined in OGIP standard
        with numpyro.plate("obs_plate_" + name, len(obs.folded_counts)):
            numpyro.sample(
                "obs_" + name,
                Poisson(countrate + bkg_countrate / obs.folded_backratio.data),
                obs=obs.folded_counts.data if observed else None,
            )

    return numpyro_model


def build_prior(prior: 'PriorDictType', expand_shape: tuple = (), prefix=""):
    """
    Transform a dictionary of prior distributions into a dictionary of parameters sampled from the prior.
    Must be used within a numpyro model.
    """
    parameters = dict(hk.data_structures.to_haiku_dict(prior))

    for i, (m, n, sample) in enumerate(hk.data_structures.traverse(prior)):
        if isinstance(sample, Distribution):
            parameters[m][n] = jnp.ones(expand_shape) * numpyro.sample(f"{prefix}{m}_{n}", sample)

        elif isinstance(sample, ArrayLike):
            parameters[m][n] = jnp.ones(expand_shape) * sample

        else:
            raise ValueError(
                f"Invalid prior type {type(sample)} for parameter {prefix}{m}_{n} : {sample}"
            )

    return parameters


from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar

import arviz as az
import astropy.cosmology.units as cu
import astropy.units as u
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from astropy.cosmology import Cosmology, Planck18
from astropy.units import Unit
from chainconsumer import Chain, ChainConsumer, PlotConfig
from haiku.data_structures import traverse
from jax.typing import ArrayLike
from numpyro.handlers import seed
from scipy.integrate import trapezoid
from scipy.special import gammaln
from scipy.stats import nbinom

if TYPE_CHECKING:
    from ..fit import BayesianModel
    from ..model.background import BackgroundModel

K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")


class HaikuDict(dict[str, dict[str, T]]): ...


def _plot_binned_samples_with_error(
    ax: plt.Axes,
    x_bins: ArrayLike,
    denominator: ArrayLike | None = None,
    y_samples: ArrayLike | None = None,
    color=(0.15, 0.25, 0.45),
    percentile: tuple = (16, 84),
):
    """
    Helper function to plot the posterior predictive distribution of the model. The function
    computes the percentiles of the posterior predictive distribution and plot them as a shaded
    area. If the observed data is provided, it is also plotted as a step function.

    Parameters
    ----------
        x_bins: The bin edges of the data (2 x N).
        y_samples: The samples of the posterior predictive distribution (Samples X N).
        denominator: Values used to divided the samples, i.e. to get energy flux (N).
        ax: The matplotlib axes object.
        color: The color of the posterior predictive distribution.
        y_observed: The observed data (N).
        label: The label of the observed data.
        percentile: The percentile of the posterior predictive distribution to plot.
    """

    mean, envelope = None, None

    if denominator is None:
        denominator = np.ones_like(x_bins[0])

    mean = ax.stairs(
        list(np.median(y_samples, axis=0) / denominator),
        edges=[*list(x_bins[0]), x_bins[1][-1]],
        color=color,
        alpha=0.7,
    )

    if y_samples is not None:
        if denominator is None:
            denominator = np.ones_like(x_bins[0])

        percentiles = np.percentile(y_samples, percentile, axis=0)

        # The legend cannot handle fill_between, so we pass a fill to get a fancy icon
        (envelope,) = ax.fill(np.nan, np.nan, alpha=0.3, facecolor=color)

        ax.stairs(
            percentiles[1] / denominator,
            edges=[*list(x_bins[0]), x_bins[1][-1]],
            baseline=percentiles[0] / denominator,
            alpha=0.3,
            fill=True,
            color=color,
        )

    return [(mean, envelope)]


class FitResult:
    """
    Container for the result of a fit using any ModelFitter class.
    """

    # TODO : Add type hints
    def __init__(
        self,
        bayesian_fitter: BayesianModel,
        inference_data: az.InferenceData,
        structure: Mapping[K, V],
        background_model: BackgroundModel = None,
    ):
        self.model = bayesian_fitter.model
        self.bayesian_fitter = bayesian_fitter
        self.inference_data = inference_data
        self.obsconfs = bayesian_fitter.observation_container
        self.background_model = background_model
        self._structure = structure

        # Add the model used in fit to the metadata
        for group in self.inference_data.groups():
            group_name = group.split("/")[-1]
            metadata = getattr(self.inference_data, group_name).attrs
            metadata["model"] = str(self.model)
            # TODO : Store metadata about observations used in the fitting process

    @property
    def converged(self) -> bool:
        r"""
        Convergence of the chain as computed by the $\hat{R}$ statistic.
        """

        return all(az.rhat(self.inference_data) < 1.01)

    @property
    def _structured_samples(self):
        """
        Get samples from the parameter posterior distribution but keep their shape in terms of draw and chains.
        """

        samples_flat = self._structured_samples_flat

        samples_haiku = {}

        for module, parameter, value in traverse(self._structure):
            if samples_haiku.get(module, None) is None:
                samples_haiku[module] = {}
            samples_haiku[module][parameter] = samples_flat[f"{module}_{parameter}"]

        return samples_haiku

    @property
    def _structured_samples_flat(self):
        """
        Get samples from the parameter posterior distribution but keep their shape in terms of draw and chains.
        """

        var_names = [f"{m}_{n}" for m, n, _ in traverse(self._structure)]
        posterior = az.extract(self.inference_data, var_names=var_names, combined=False)
        samples_flat = {key: posterior[key].data for key in var_names}

        return samples_flat

    @property
    def input_parameters(self) -> HaikuDict[ArrayLike]:
        """
        The input parameters of the model.
        """

        posterior = az.extract(self.inference_data, combined=False)

        samples_shape = (len(posterior.coords["chain"]), len(posterior.coords["draw"]))

        total_shape = tuple(posterior.sizes[d] for d in posterior.coords)

        posterior = {key: posterior[key].data for key in posterior.data_vars}

        with seed(rng_seed=0):
            input_parameters = self.bayesian_fitter.prior_distributions_func()

        for module, parameter, value in traverse(input_parameters):
            if f"{module}_{parameter}" in posterior.keys():
                # We add as extra dimension as there might be different values per observation
                if posterior[f"{module}_{parameter}"].shape == samples_shape:
                    to_set = posterior[f"{module}_{parameter}"][..., None]
                else:
                    to_set = posterior[f"{module}_{parameter}"]

                input_parameters[module][parameter] = to_set

            else:
                # The parameter is fixed in this case, so we just broadcast is over chain and draws
                input_parameters[module][parameter] = value[None, None, ...]

            if len(total_shape) < len(input_parameters[module][parameter].shape):
                # If there are only chains and draws, we reduce
                input_parameters[module][parameter] = input_parameters[module][parameter][..., 0]

            else:
                input_parameters[module][parameter] = jnp.broadcast_to(
                    input_parameters[module][parameter], total_shape
                )

        return input_parameters

    def photon_flux(
        self,
        e_min: float,
        e_max: float,
        unit: Unit = u.photon / u.cm**2 / u.s,
        register: bool = False,
    ) -> ArrayLike:
        """
        Compute the unfolded photon flux in a given energy band. The flux is then added to
        the result parameters so covariance can be plotted.

        Parameters:
            e_min: The lower bound of the energy band in observer frame.
            e_max: The upper bound of the energy band in observer frame.
            unit: The unit of the photon flux.
            register: Whether to register the flux with the other posterior parameters.

        !!! warning
            Computation of the folded flux is not implemented yet. Feel free to open an
            [issue](https://github.com/renecotyfanboy/jaxspec/issues) in the GitHub repository.
        """

        @jax.jit
        @jnp.vectorize
        def vectorized_flux(*pars):
            parameters_pytree = jax.tree.unflatten(pytree_def, pars)
            return self.model.photon_flux(
                parameters_pytree, jnp.asarray([e_min]), jnp.asarray([e_max]), n_points=100
            )[0]

        flat_tree, pytree_def = jax.tree.flatten(self.input_parameters)
        flux = vectorized_flux(*flat_tree)
        conversion_factor = (u.photon / u.cm**2 / u.s).to(unit)
        value = flux * conversion_factor

        if register:
            self.inference_data.posterior[f"photon_flux_{e_min:.1f}_{e_max:.1f}"] = (
                list(self.inference_data.posterior.coords),
                value,
            )

        return value

    def energy_flux(
        self,
        e_min: float,
        e_max: float,
        unit: Unit = u.erg / u.cm**2 / u.s,
        register: bool = False,
    ) -> ArrayLike:
        """
        Compute the unfolded energy flux in a given energy band. The flux is then added to
        the result parameters so covariance can be plotted.

        Parameters:
            e_min: The lower bound of the energy band in observer frame.
            e_max: The upper bound of the energy band in observer frame.
            unit: The unit of the energy flux.
            register: Whether to register the flux with the other posterior parameters.

        !!! warning
            Computation of the folded flux is not implemented yet. Feel free to open an
            [issue](https://github.com/renecotyfanboy/jaxspec/issues) in the GitHub repository.
        """

        @jax.jit
        @jnp.vectorize
        def vectorized_flux(*pars):
            parameters_pytree = jax.tree.unflatten(pytree_def, pars)
            return self.model.energy_flux(
                parameters_pytree, jnp.asarray([e_min]), jnp.asarray([e_max]), n_points=100
            )[0]

        flat_tree, pytree_def = jax.tree.flatten(self.input_parameters)
        flux = vectorized_flux(*flat_tree)
        conversion_factor = (u.keV / u.cm**2 / u.s).to(unit)
        value = flux * conversion_factor

        if register:
            self.inference_data.posterior[f"energy_flux_{e_min:.1f}_{e_max:.1f}"] = (
                list(self.inference_data.posterior.coords),
                value,
            )

        return value

    def luminosity(
        self,
        e_min: float,
        e_max: float,
        redshift: float | ArrayLike = None,
        distance: float | ArrayLike = None,
        observer_frame: bool = True,
        cosmology: Cosmology = Planck18,
        unit: Unit = u.erg / u.s,
        register: bool = False,
    ) -> ArrayLike:
        """
        Compute the luminosity of the source specifying its redshift. The luminosity is then added to
        the result parameters so covariance can be plotted.

        Parameters:
            e_min: The lower bound of the energy band.
            e_max: The upper bound of the energy band.
            redshift: The redshift of the source. It can be a distribution of redshifts.
            observer_frame: Whether the input bands are defined in observer frame or not.
            cosmology: Chosen cosmology.
            unit: The unit of the luminosity.
            register: Whether to register the flux with the other posterior parameters.
        """

        if not observer_frame:
            raise NotImplementedError()

        if redshift is None and distance is None:
            raise ValueError("Either redshift or distance must be specified.")

        if distance is not None:
            if redshift is not None:
                raise ValueError("Redshift must be None as a distance is specified.")
            else:
                redshift = distance.to(
                    cu.redshift, cu.redshift_distance(cosmology, kind="luminosity")
                ).value

        @jax.jit
        @jnp.vectorize
        def vectorized_flux(*pars):
            parameters_pytree = jax.tree.unflatten(pytree_def, pars)
            return self.model.energy_flux(
                parameters_pytree,
                jnp.asarray([e_min]) * (1 + redshift),
                jnp.asarray([e_max]) * (1 + redshift),
                n_points=100,
            )[0]

        flat_tree, pytree_def = jax.tree.flatten(self.input_parameters)
        flux = vectorized_flux(*flat_tree) * (u.keV / u.cm**2 / u.s)
        value = (flux * (4 * np.pi * cosmology.luminosity_distance(redshift) ** 2)).to(unit)

        if register:
            self.inference_data.posterior[f"luminosity_{e_min:.1f}_{e_max:.1f}"] = (
                list(self.inference_data.posterior.coords),
                value,
            )

        return value

    def to_chain(self, name: str, parameters_type: Literal["model", "bkg"] = "model") -> Chain:
        """
        Return a ChainConsumer Chain object from the posterior distribution of the parameters_type.

        Parameters:
            name: The name of the chain.
            parameters_type: The parameters_type to include in the chain.
        """

        obs_id = self.inference_data.copy()

        if parameters_type == "model":
            keys_to_drop = [
                key
                for key in obs_id.posterior.keys()
                if (key.startswith("_") or key.startswith("bkg"))
            ]
        elif parameters_type == "bkg":
            keys_to_drop = [key for key in obs_id.posterior.keys() if not key.startswith("bkg")]
        else:
            raise ValueError(f"Unknown value for parameters_type: {parameters_type}")

        obs_id.posterior = obs_id.posterior.drop_vars(keys_to_drop)
        chain = Chain.from_arviz(obs_id, name)

        """
        chain.samples.columns = [
            format_parameters(parameter) for parameter in chain.samples.columns
        ]
        """

        return chain

    @property
    def samples_haiku(self) -> HaikuDict[ArrayLike]:
        """
        Haiku-like structure for the samples e.g.

        ```
        {
            'powerlaw_1' :
            {
                'alpha': ...,
                'amplitude': ...
            },

            'blackbody_1':
            {
                'kT': ...,
                'norm': ...
            },

            'tbabs_1':
            {
                'nH': ...
            }
        }
        ```

        """

        params = {}

        for module, parameter, value in traverse(self._structure):
            if params.get(module, None) is None:
                params[module] = {}
            params[module][parameter] = self.samples_flat[f"{module}_{parameter}"]

        return params

    @property
    def samples_flat(self) -> dict[str, ArrayLike]:
        """
        Flat structure for the samples e.g.

        ```
        {
            'powerlaw_1_alpha': ...,
            'powerlaw_1_amplitude': ...,
            'blackbody_1_kT': ...,
            'blackbody_1_norm': ...,
            'tbabs_1_nH': ...,
        }
        ```
        """
        var_names = [f"{m}_{n}" for m, n, _ in traverse(self._structure)]
        posterior = az.extract(self.inference_data, var_names=var_names)
        return {key: posterior[key].data for key in var_names}

    @property
    def log_likelihood(self) -> xr.Dataset:
        """
        Return the log_likelihood of each observation
        """
        log_likelihood = az.extract(self.inference_data, group="log_likelihood")
        dimensions_to_reduce = [
            coord for coord in log_likelihood.coords if coord not in ["sample", "draw", "chain"]
        ]
        return log_likelihood.sum(dimensions_to_reduce)

    @property
    def c_stat(self):
        r"""
        Return the C-statistic of the model

        The C-statistic is defined as:

        $$ C = 2 \sum_{i} M - D*log(M) + D*log(D) - D $$
        or
        $$ C = 2 \sum_{i} M - D*log(M)$$
        for bins with no counts

        """

        exclude_dims = ["chain", "draw", "sample"]
        all_dims = list(self.inference_data.log_likelihood.dims)
        reduce_dims = [dim for dim in all_dims if dim not in exclude_dims]
        data = self.inference_data.observed_data
        c_stat = -2 * (
            self.log_likelihood
            + (gammaln(data + 1) - (xr.where(data > 0, data * (np.log(data) - 1), 0))).sum(
                dim=reduce_dims
            )
        )

        return c_stat

    def plot_ppc(
        self,
        percentile: tuple[int, int] = (16, 84),
        x_unit: str | u.Unit = "keV",
        y_type: Literal[
            "counts", "countrate", "photon_flux", "photon_flux_density"
        ] = "photon_flux_density",
    ) -> plt.Figure:
        r"""
        Plot the posterior predictive distribution of the model. It also features a residual plot, defined using the
        following formula:

        $$ \text{Residual} = \frac{\text{Observed counts} - \text{Posterior counts}}
        {(\text{Posterior counts})_{84\%}-(\text{Posterior counts})_{16\%}} $$

        Parameters:
            percentile: The percentile of the posterior predictive distribution to plot.
            x_unit: The units of the x-axis. It can be either a string (parsable by astropy.units) or an astropy unit. It must be homogeneous to either a length, a frequency or an energy.
            y_type: The type of the y-axis. It can be either "counts", "countrate", "photon_flux" or "photon_flux_density".

        Returns:
            The matplotlib figure.
        """

        obsconf_container = self.obsconfs
        x_unit = u.Unit(x_unit)

        match y_type:
            case "counts":
                y_units = u.ct
            case "countrate":
                y_units = u.ct / u.s
            case "photon_flux":
                y_units = u.ct / u.cm**2 / u.s
            case "photon_flux_density":
                y_units = u.ct / u.cm**2 / u.s / x_unit
            case _:
                raise ValueError(
                    f"Unknown y_type: {y_type}. Must be 'counts', 'countrate', 'photon_flux' or 'photon_flux_density'"
                )

        color = (0.15, 0.25, 0.45)

        with plt.style.context("default"):
            # Note to Simon : do not change xbins[1] - xbins[0] to
            # np.diff, you already did this twice and forgot that it does not work since diff keeps the dimensions
            # and enable weird broadcasting that makes the plot fail

            fig, axs = plt.subplots(
                2,
                len(obsconf_container),
                figsize=(6 * len(obsconf_container), 6),
                sharex=True,
                height_ratios=[0.7, 0.3],
            )

            plot_ylabels_once = True

            for name, obsconf, ax in zip(
                obsconf_container.keys(),
                obsconf_container.values(),
                axs.T if len(obsconf_container) > 1 else [axs],
            ):
                legend_plots = []
                legend_labels = []
                count = az.extract(
                    self.inference_data, var_names=f"obs_{name}", group="posterior_predictive"
                ).values.T
                bkg_count = (
                    None
                    if self.background_model is None
                    else az.extract(
                        self.inference_data, var_names=f"bkg_{name}", group="posterior_predictive"
                    ).values.T
                )

                xbins = obsconf.out_energies * u.keV
                xbins = xbins.to(x_unit, u.spectral())

                # This compute the total effective area within all bins
                # This is a bit weird since the following computation is equivalent to ignoring the RMF
                exposure = obsconf.exposure.data * u.s
                mid_bins_arf = obsconf.in_energies.mean(axis=0) * u.keV
                mid_bins_arf = mid_bins_arf.to(x_unit, u.spectral())
                e_grid = np.linspace(*xbins, 10)
                interpolated_arf = np.interp(e_grid, mid_bins_arf, obsconf.area)
                integrated_arf = (
                    trapezoid(interpolated_arf, x=e_grid, axis=0)
                    / (
                        np.abs(
                            xbins[1] - xbins[0]
                        )  # Must fold in abs because some units reverse the ordering of the bins
                    )
                    * u.cm**2
                )

                match y_type:
                    case "counts":
                        denominator = 1
                    case "countrate":
                        denominator = exposure
                    case "photon_flux":
                        denominator = integrated_arf * exposure
                    case "photon_flux_density":
                        denominator = (xbins[1] - xbins[0]) * integrated_arf * exposure

                y_samples = (count * u.ct / denominator).to(y_units)
                y_observed = (obsconf.folded_counts.data * u.ct / denominator).to(y_units)
                y_observed_low = (
                    nbinom.ppf(percentile[0] / 100, obsconf.folded_counts.data, 0.5)
                    * u.ct
                    / denominator
                ).to(y_units)
                y_observed_high = (
                    nbinom.ppf(percentile[1] / 100, obsconf.folded_counts.data, 0.5)
                    * u.ct
                    / denominator
                ).to(y_units)

                # Use the helper function to plot the data and posterior predictive
                legend_plots += _plot_binned_samples_with_error(
                    ax[0],
                    xbins.value,
                    y_samples=y_samples.value,
                    denominator=np.ones_like(y_observed).value,
                    color=color,
                    percentile=percentile,
                )

                legend_labels.append("Model")

                true_data_plot = ax[0].errorbar(
                    np.sqrt(xbins.value[0] * xbins.value[1]),
                    y_observed.value,
                    xerr=np.abs(xbins.value - np.sqrt(xbins.value[0] * xbins.value[1])),
                    yerr=[
                        y_observed.value - y_observed_low.value,
                        y_observed_high.value - y_observed.value,
                    ],
                    color="black",
                    linestyle="none",
                    alpha=0.3,
                    capsize=2,
                )

                legend_plots.append((true_data_plot,))
                legend_labels.append("Observed")

                if self.background_model is not None:
                    # We plot the background only if it is included in the fit, i.e. by subtracting
                    ratio = obsconf.folded_backratio.data
                    y_samples_bkg = (bkg_count * u.ct / (denominator * ratio)).to(y_units)
                    y_observed_bkg = (
                        obsconf.folded_background.data * u.ct / (denominator * ratio)
                    ).to(y_units)
                    y_observed_bkg_low = (
                        nbinom.ppf(percentile[0] / 100, obsconf.folded_background.data, 0.5)
                        * u.ct
                        / (denominator * ratio)
                    ).to(y_units)
                    y_observed_bkg_high = (
                        nbinom.ppf(percentile[1] / 100, obsconf.folded_background.data, 0.5)
                        * u.ct
                        / (denominator * ratio)
                    ).to(y_units)

                    legend_plots += _plot_binned_samples_with_error(
                        ax[0],
                        xbins.value,
                        y_samples=y_samples_bkg.value,
                        denominator=np.ones_like(y_observed).value,
                        color=(0.26787604, 0.60085972, 0.63302651),
                        percentile=percentile,
                    )

                    legend_labels.append("Model (bkg)")

                    true_bkg_plot = ax[0].errorbar(
                        np.sqrt(xbins.value[0] * xbins.value[1]),
                        y_observed_bkg.value,
                        xerr=np.abs(xbins.value - np.sqrt(xbins.value[0] * xbins.value[1])),
                        yerr=[
                            y_observed_bkg.value - y_observed_bkg_low.value,
                            y_observed_bkg_high.value - y_observed_bkg.value,
                        ],
                        color="black",
                        linestyle="none",
                        alpha=0.3,
                        capsize=2,
                    )

                    legend_plots.append((true_bkg_plot,))
                    legend_labels.append("Observed (bkg)")

                residual_samples = (obsconf.folded_counts.data - count) / np.diff(
                    np.percentile(count, percentile, axis=0), axis=0
                )

                residuals = np.percentile(
                    residual_samples,
                    percentile,
                    axis=0,
                )

                median_residuals = np.median(
                    residual_samples,
                    axis=0,
                )

                ax[1].stairs(
                    residuals[1],
                    edges=[*list(xbins.value[0]), xbins.value[1][-1]],
                    baseline=list(residuals[0]),
                    alpha=0.3,
                    facecolor=color,
                    fill=True,
                )

                ax[1].stairs(
                    median_residuals,
                    edges=[*list(xbins.value[0]), xbins.value[1][-1]],
                    color=color,
                    alpha=0.7,
                )

                max_residuals = np.max(np.abs(residuals))

                ax[0].loglog()
                ax[1].set_ylim(-max(3.5, max_residuals), +max(3.5, max_residuals))

                if plot_ylabels_once:
                    ax[0].set_ylabel(f"Folded spectrum\n [{y_units:latex_inline}]")
                    ax[1].set_ylabel("Residuals \n" + r"[$\sigma$]")
                    plot_ylabels_once = False

                match getattr(x_unit, "physical_type"):
                    case "length":
                        ax[1].set_xlabel(f"Wavelength \n[{x_unit:latex_inline}]")
                    case "energy":
                        ax[1].set_xlabel(f"Energy \n[{x_unit:latex_inline}]")
                    case "frequency":
                        ax[1].set_xlabel(f"Frequency \n[{x_unit:latex_inline}]")
                    case _:
                        RuntimeError(
                            f"Unknown physical type for x_units: {x_unit}. "
                            f"Must be 'length', 'energy' or 'frequency'"
                        )

                ax[1].axhline(0, color=color, ls="--")
                ax[1].axhline(-3, color=color, ls=":")
                ax[1].axhline(3, color=color, ls=":")

                # ax[1].set_xticks(xticks, labels=xticks)
                # ax[1].xaxis.set_minor_formatter(ticker.LogFormatter(minor_thresholds=(np.inf, np.inf)))
                ax[1].set_yticks([-3, 0, 3], labels=[-3, 0, 3])
                ax[1].set_yticks(range(-3, 4), minor=True)

                ax[0].set_xlim(xbins.value.min(), xbins.value.max())

                ax[0].legend(legend_plots, legend_labels)
                fig.suptitle(self.model.to_string())
                fig.align_ylabels()
                plt.subplots_adjust(hspace=0.0)
                fig.tight_layout()

            return fig

    def table(self) -> str:
        r"""
        Return a formatted $\LaTeX$ table of the results of the fit.
        """

        consumer = ChainConsumer()
        consumer.add_chain(self.to_chain(self.model.to_string()))

        return consumer.analysis.get_latex_table(caption="Results of the fit", label="tab:results")

    def plot_corner(
        self,
        config: PlotConfig = PlotConfig(usetex=False, summarise=False, label_font_size=12),
        **kwargs: Any,
    ) -> plt.Figure:
        """
        Plot the corner plot of the posterior distribution of the parameters_type. This method uses the ChainConsumer.

        Parameters:
            config: The configuration of the plot.
            **kwargs: Additional arguments passed to ChainConsumer.plotter.plot. Some useful parameters are :
                - columns : list of parameters to plot.
        """

        consumer = ChainConsumer()
        consumer.add_chain(self.to_chain(self.model.to_string()))
        consumer.set_plot_config(config)

        # Context for default mpl style
        with plt.style.context("default"):
            return consumer.plotter.plot(**kwargs)
